import matplotlib.pylab as plt
import numpy as np
import xarray as xr
from xarray import DataArray

from indica.defaults.load_defaults import load_default_objects
from indica.models import ThomsonScattering
from indica.operators import tomo_1D
from indica.operators.centrifugal_asymmetry import centrifugal_asymmetry_2d_map
from indica.operators.centrifugal_asymmetry import centrifugal_asymmetry_parameter
from indica.operators.spline_fit_R_shift import fit_profile_and_R_shift
from indica.operators.tomo_asymmetry import InvertPoloidalAsymmetry
from indica.readers.modelreader import ModelReader
from indica.utilities import set_axis_sci
from indica.utilities import set_plot_colors

PLASMA = load_default_objects("st40", "plasma")
EQUILIBRIUM = load_default_objects("st40", "equilibrium")
TRANSFORMS = load_default_objects("st40", "geometry")
PLASMA.set_equilibrium(EQUILIBRIUM)

NPLOT = 3
CM, COLS = set_plot_colors()


def example_poloidal_asymmetry():
    asymmetry_parameter = centrifugal_asymmetry_parameter(
        PLASMA.ion_density,
        PLASMA.ion_temperature,
        PLASMA.electron_temperature,
        PLASMA.toroidal_rotation,
        PLASMA.meanz,
        PLASMA.zeff,
        PLASMA.main_ion,
    )

    ion_density_2d = centrifugal_asymmetry_2d_map(
        PLASMA.ion_density,
        asymmetry_parameter,
        EQUILIBRIUM,
    )

    return ion_density_2d


def example_tomo_asymmetry(
    instrument: str = "sxrc_xy1",
    asymmetric_profile: bool = True,
    plot: bool = True,
    element: str = "ar",
):

    if asymmetric_profile:
        ion_density_2d = example_poloidal_asymmetry()
    else:
        rho_2d = PLASMA.equilibrium.rhop.interp(t=PLASMA.t)
        ion_density_2d = PLASMA.ion_density.interp(rhop=rho_2d)

    imp_density_2d = ion_density_2d.sel(element=element).drop_vars("element")
    el_density_2d = PLASMA.electron_density.interp(rhop=imp_density_2d.rhop)
    lz_tot_2d = (
        PLASMA.lz_tot[element].sum("ion_charge").interp(rhop=imp_density_2d.rhop)
    )
    phantom_emission = el_density_2d * imp_density_2d * lz_tot_2d
    los_transform = TRANSFORMS[instrument]
    los_transform.set_equilibrium(EQUILIBRIUM)
    los_integral = los_transform.integrate_on_los(
        phantom_emission, phantom_emission.t.values
    )

    print("\n Running asymmetry inference...")
    invert_asymm = InvertPoloidalAsymmetry()
    bckc_emission, bckc, profile, asymmetry = invert_asymm(
        los_integral,
        los_transform,
    )
    print("...and finished \n")

    if plot:
        plt.ioff()
        los_transform.plot()
        plt.show(block=True)

        for i, t in enumerate(bckc.t):
            if i % np.floor(np.size(bckc.t) / NPLOT) != 0:
                continue
            kwargs_phantom = dict(
                marker="o",
                label=f"Phantom @ t={int(t*1.e3)} ms",
                color="r",
                zorder=1,
                alpha=0.8,
            )
            kwargs_bckc = dict(
                marker="x",
                label=f"Back-calculated @ t={int(t*1.e3)} ms",
                color="b",
                zorder=2,
                alpha=0.8,
            )
            plt.figure()
            plt.errorbar(
                invert_asymm.data.channel,
                invert_asymm.data.sel(t=t, method="nearest"),
                invert_asymm.error.sel(t=t, method="nearest"),
                **kwargs_phantom,
            )
            bckc.attrs = {"long_name": "Brightness", "units": "W/m$^2$"}
            bckc.sel(t=t).plot(**kwargs_bckc)
            plt.title(f"LOS integrals @ t={int(t*1.e3)} ms")
            plt.legend()
            set_axis_sci()

            plt.figure()
            phantom_emission.attrs = {"long_name": "Emissivity", "units": "W/m$^3$"}
            phantom_emission.sel(t=t).plot()
            plt.title(f"2D Phantom @ t={int(t*1.e3)} ms")
            plt.axis("equal")

            # plt.figure()
            # residuals = (bckc_emission - phantom_emission)/phantom_emission * 100
            # residuals.attrs = {"long_name": "(Phantom - Bckc)/Phantom", "units": "%"}
            # residuals = xr.where(residuals < 100, residuals, np.nan)
            # residuals.sel(t=t).plot()
            # plt.title("Phantom - Bckc normalised residuals")
            # plt.axis("equal")

            plt.figure()
            kwargs_phantom = dict(
                label="Phantom",
                color="r",
                zorder=1,
                alpha=0.8,
                linestyle="dashed",
            )
            kwargs_bckc = dict(
                label="Back-calculated",
                color="b",
                zorder=2,
                alpha=0.8,
            )
            if asymmetric_profile:
                phantom_emission.sel(t=t).sel(z=0, method="nearest").plot(
                    **kwargs_phantom,
                )
                bckc_emission.sel(t=t).sel(z=0, method="nearest").plot(
                    **kwargs_bckc,
                )
            else:
                rmag = PLASMA.equilibrium.rmag.sel(t=t, method="nearest")
                rho_lfs = (
                    PLASMA.equilibrium.rhop.sel(t=t, method="nearest")
                    .sel(z=0, method="nearest")
                    .sel(R=slice(rmag - 0.01, 1))
                )
                profile_1d = (
                    (
                        phantom_emission.sel(t=t)
                        .sel(z=0, method="nearest")
                        .interp(R=rho_lfs.R)
                    )
                    .assign_coords(rhop=("R", rho_lfs.data))
                    .swap_dims({"R": "rhop"})
                )
                profile_1d_bckc = (
                    (
                        bckc_emission.sel(t=t)
                        .sel(z=0, method="nearest")
                        .interp(R=rho_lfs.R)
                    )
                    .assign_coords(rhop=("R", rho_lfs.data))
                    .swap_dims({"R": "rhop"})
                )
                profile_1d.plot(
                    **kwargs_phantom,
                )
                profile_1d_bckc.plot(
                    **kwargs_bckc,
                )
            plt.title(f"Midplane cut (z=0) @ t={int(t*1.e3)} ms")
            set_axis_sci()
            plt.legend()
            plt.show()


def example_tomo_1D(
    instrument: str = "sxrc_xy1",
    asymmetric_profile: bool = False,
    element: str = "ar",
    reg_level_guess: float = 0.8,
    plot: bool = True,
):

    if asymmetric_profile:
        ion_density_2d = example_poloidal_asymmetry()
        emissivity = None
    else:
        rho_2d = PLASMA.equilibrium.rhop.interp(t=PLASMA.t.values)
        ion_density_2d = PLASMA.ion_density.interp(rhop=rho_2d)
        emissivity = (
            PLASMA.electron_density
            * PLASMA.ion_density.sel(element=element).drop_vars("element")
            * PLASMA.lz_tot[element].sum("ion_charge")
        )

    imp_density_2d = ion_density_2d.sel(element=element).drop_vars("element")
    el_density_2d = PLASMA.electron_density.interp(rhop=imp_density_2d.rhop)
    lz_tot_2d = (
        PLASMA.lz_tot[element].sum("ion_charge").interp(rhop=imp_density_2d.rhop)
    )
    phantom_emission = el_density_2d * imp_density_2d * lz_tot_2d
    los_transform = TRANSFORMS[instrument]
    los_transform.set_equilibrium(PLASMA.equilibrium)
    los_integral = los_transform.integrate_on_los(
        phantom_emission, phantom_emission.t.values
    )

    z = los_transform.z.mean("beamlet")
    R = los_transform.R.mean("beamlet")
    dl = los_transform.dl
    has_data = np.logical_not(np.isnan(los_integral.isel(t=0).data))
    rho_equil = los_transform.equilibrium.rhop.interp(t=los_integral.t)
    input_dict = dict(
        brightness=los_integral.data.T,
        dl=dl,
        t=los_integral.t.data,
        R=R.data,
        z=z.data,
        rho_equil=dict(
            R=rho_equil.R.data,
            z=rho_equil.z.data,
            t=rho_equil.t.data,
            rho=rho_equil.data,
        ),
        has_data=has_data,
        debug=False,
    )
    if emissivity is not None:
        input_dict["emissivity"] = emissivity

    tomo = tomo_1D.SXR_tomography(input_dict, reg_level_guess=reg_level_guess)
    tomo()

    inverted_emissivity = DataArray(
        tomo.emiss, coords=[("t", tomo.tvec), ("rhop", tomo.rho_grid_centers)]
    )
    inverted_error = DataArray(
        tomo.emiss_err,
        coords=[("t", tomo.tvec), ("rhop", tomo.rho_grid_centers)],
    )
    inverted_emissivity = inverted_emissivity.assign_coords(
        error=(inverted_emissivity.dims, inverted_error.data)
    )

    data_tomo = los_integral
    bckc_tomo = DataArray(tomo.backprojection.T, coords=data_tomo.coords)

    if plot:
        plt.ioff()
        los_transform.plot()
        tomo.show_reconstruction()
        plt.show()

    return inverted_emissivity, data_tomo, bckc_tomo


def example_fit_ts(
    fit_R_shift: bool = False,
    verbose: bool = False,
    plot: bool = True,
):

    models = {"ts": ThomsonScattering}
    model_reader = ModelReader(models)
    model_reader.set_plasma(PLASMA)
    model_reader.set_geometry_transforms(TRANSFORMS, EQUILIBRIUM)
    bckc = model_reader()

    te_data = bckc["ts"]["te"]
    te_err = bckc["ts"]["te"].error
    ne_data = bckc["ts"]["ne"]
    ne_err = bckc["ts"]["ne"].error

    if not fit_R_shift:
        R_shift = xr.full_like(te_data.t, 0.0)
    else:
        R_shift = None

    ts_R = te_data.transform.R
    ts_z = te_data.transform.z
    equilibrium = te_data.transform.equilibrium
    te_fit, te_R_shift, te_rho = fit_profile_and_R_shift(
        ts_R,
        ts_z,
        te_data,
        te_err,
        xknots=[0, 0.4, 0.85, 0.9, 0.98, 1.1],
        equilibrium=equilibrium,
        R_shift=R_shift,
        verbose=verbose,
    )
    ne_fit, ne_R_shift, ne_rho = fit_profile_and_R_shift(
        ts_R,
        ts_z,
        ne_data,
        ne_err,
        xknots=[0, 0.4, 0.85, 0.9, 0.98, 1.1],
        equilibrium=equilibrium,
        R_shift=te_R_shift,
        verbose=verbose,
    )
    te_fit.attrs["R_shift"] = te_R_shift
    ne_fit.attrs["R_shift"] = ne_R_shift

    if plot:
        time = te_data.t
        cols = CM(np.linspace(0.1, 0.75, len(time), dtype=float))

        plt.ioff()
        plt.figure()
        _R_shift = [f"{(ne_fit.R_shift.sel(t=t).values * 100):.1f}" for t in time]
        for i, t in enumerate(time.values):
            if i % NPLOT:
                continue
            plt.errorbar(
                ne_data.transform.rhop.sel(t=t),
                ne_data.sel(t=t),
                ne_data.error.sel(t=t),
                color=cols[i],
                marker="o",
                label=rf"t={int(t*1.e3)} ms $\delta$R={_R_shift[i]} cm",
                alpha=0.6,
            )
            ne_fit.sel(t=t).plot(color=cols[i], linewidth=4, zorder=0)
        plt.ylabel("Ne (m${-3}$)")
        plt.xlabel("Rho-poloidal")
        plt.title("TS Ne data & fits")
        plt.xlim(0, 1.1)
        plt.ylim(0, np.nanmax(ne_fit) * 1.2)
        plt.legend()

        plt.figure()
        for i, t in enumerate(time.values):
            if i % NPLOT:
                continue
            plt.errorbar(
                te_data.transform.rhop.sel(t=t),
                te_data.sel(t=t),
                te_data.error.sel(t=t),
                color=cols[i],
                marker="o",
                label=rf"t={int(t*1.e3)} ms $\delta$R={_R_shift[i]} cm",
                alpha=0.6,
            )
            te_fit.sel(t=t).plot(color=cols[i], linewidth=4, zorder=0)
        plt.ylabel("Te (eV)")
        plt.xlabel("Rho-poloidal")
        plt.title("TS Te data & fits")
        plt.xlim(0, 1.1)
        plt.ylim(0, np.nanmax(te_fit) * 1.2)
        plt.legend()
        plt.show()

    return te_data, ne_data, te_fit, ne_fit



import pathlib
import subprocess

try:
    git_hash = subprocess.check_output(
        ["git", "describe", "--always"], text=True
    ).strip()
    git_diff = subprocess.check_output(
        ["git", "diff", "HEAD", "--", "indica"], text=True
    ).strip()
    if len(git_diff) > 0:
        git_hash += "-dirty"
except FileNotFoundError:
    git_hash = "UNKNOWN"
with open(pathlib.Path("indica") / "git_version", "w") as f:
    f.write(git_hash)



from abc import ABC
from abc import abstractmethod
from enum import Enum
from typing import Callable

import numpy as np
from scipy.stats import gaussian_kde
from scipy.stats import loguniform
from scipy.stats import uniform


def greater_than(x1, x2):
    return np.where(x1 > x2, 1, 0)


def get_uniform(lower, upper):
    # Less confusing parametrisation of scipy.stats uniform
    return uniform(loc=lower, scale=upper - lower)


class PriorType(Enum):
    """
    PriorType differentiates between the below prior types.
    BASIC: 1D PDF
    COND: generic relationship between 2+ parameters of form: func(*parameters)->float
    COMPOUND: ND PDF
    """

    BASIC = 1
    COND = 2
    COMPOUND = 3


class Prior(ABC):
    def __init__(
        self,
        prior_func: Callable = None,
        labels: tuple = None,
        type: PriorType = None,
    ):
        self.prior_func = prior_func
        self.labels = (
            labels  # to identify mapping between prior names and ndim prior funcs
        )
        self.type = type

    @abstractmethod
    def pdf(self, value):
        return None

    @abstractmethod
    def rvs(self, size):
        return None


class PriorBasic(Prior):
    def __init__(self, prior_func: Callable = None, labels: tuple = None):
        super().__init__(prior_func=prior_func, labels=labels, type=PriorType.BASIC)

    def pdf(self, value):
        return self.prior_func.pdf(value)

    def rvs(self, size):
        return self.prior_func.rvs(size)


class PriorCond(Prior):
    """
    Generic relationship between 2 or more parameters
    """

    def __init__(
        self,
        prior_func: Callable = None,
        labels: tuple = None,
    ):
        super().__init__(prior_func=prior_func, labels=labels, type=PriorType.COND)

    def pdf(self, *values):
        return self.prior_func(*values)

    def rvs(self, size):
        return None


class PriorCompound(Prior):
    """
    ND Probability Distribution Function
    """

    def __init__(
        self,
        prior_func: gaussian_kde = None,
        labels: tuple = None,
    ):
        super().__init__(labels=labels, type=PriorType.COMPOUND)
        self.prior_func = prior_func

    def pdf(self, *values):
        _pdf = self.prior_func(np.array(values))
        if _pdf.size == 1:  # Should return scalar when evaluating one point
            _pdf = _pdf.item()
        return _pdf

    def rvs(self, size):
        return self.prior_func.resample(size).T


class PriorManager:
    def __init__(
        self,
        basic_prior_info: dict = None,
        cond_prior_info: dict = None,
    ):
        self.compound_prior_funcs = {}  # initialised later

        self.basic_prior_info = basic_prior_info
        self.cond_prior_info = cond_prior_info

        self.get_uniform = get_uniform
        self.loguniform = loguniform
        self.greater_than = greater_than

        # Initialise prior objects
        self.priors: dict = {}
        for name, prior in self.basic_prior_info.items():
            prior = getattr(self, prior[0])(*prior[1:])
            self.priors[name] = PriorBasic(prior, labels=tuple([name]))
        for name, prior in self.cond_prior_info.items():
            prior = getattr(self, prior)
            self.priors[name] = PriorCond(prior, labels=tuple(name.split("/")))

    def update_priors(self, new_priors: dict):
        #  update priors but remove all priors that match the profile names first

        prior_prefixes_to_remove = list(
            set([key.split(".")[0] for key in new_priors.keys()])
        )
        priors_to_remove = [
            prior_name
            for prior_name in self.priors.keys()
            if any(
                prior_name
                for prefix in prior_prefixes_to_remove
                if prefix in prior_name
            )
        ]
        print(f"Discarding priors: {priors_to_remove}")
        print(f"Updating with {new_priors.keys()}")
        for prior_name in priors_to_remove:
            self.priors.pop(prior_name)
        self.priors.update(new_priors)

    def get_prior_names_for_profiles(self, profile_names: list) -> list:
        #  Get all priors that correspond to the profile_names given
        prior_names = [
            prior_name
            for prior_name, prior in self.priors.items()
            for profile_name in profile_names
            if profile_name in str(prior.labels)
        ]

        return prior_names

    def get_param_names_for_profiles(self, profile_names: list) -> list:
        #  Get all parameters that correspond to the profile_names given
        param_names = [
            list(prior.labels)
            for prior_name, prior in self.priors.items()
            for profile_name in profile_names
            if profile_name in str(prior.labels)
        ]
        unpacked_names = [
            param_name
            for param_name_list in param_names
            for param_name in param_name_list
        ]
        unique_param_names = list(set(unpacked_names))
        return unique_param_names

    def ln_prior(self, parameters: dict):
        # refactor ln_prior to define interface between cond and regular priors
        return ln_prior(priors=self.priors, parameters=parameters)


def ln_prior(priors: dict, parameters: dict):
    ln_prior = 0

    for prior_name, prior in priors.items():
        param_names_in_prior = [x for x in prior.labels if x in parameters.keys()]
        if param_names_in_prior.__len__() == 0:
            # if prior assigned but no parameter then skip
            continue
        # if not all relevant params given then ignore prior
        if param_names_in_prior.__len__() != prior.labels.__len__():
            continue
        param_values = np.array([parameters[x] for x in param_names_in_prior])
        ln_prior += np.log(prior.pdf(*param_values))
    return ln_prior


def sample_from_priors(param_names: list, priors: dict, size=10) -> np.ndarray:
    samples = np.empty((param_names.__len__(), 0))
    while samples.size < param_names.__len__() * size:
        # Increase 'size * n' if too slow / looping too much
        _new_sample = {}
        for prior_name, prior in priors.items():
            if any([label not in param_names for label in prior.labels]):
                continue
            if prior.type is PriorType.BASIC:
                _new_sample[prior_name] = prior.rvs(size=size * 2)
            elif prior.type is PriorType.COMPOUND:
                tuple_sample = prior.rvs(size=size * 2)
                for idx, value in enumerate(prior.labels):
                    _new_sample[value] = tuple_sample[:, idx]
            elif prior.type is PriorType.COND:
                continue
            else:
                raise TypeError(f"{prior} is not a version of {PriorType}")

        #  Due to looping over priors the samples need to be reordered
        new_sample = {param_name: _new_sample[param_name] for param_name in param_names}
        #  Throw out samples that don't meet conditional priors and redraw
        _ln_prior = ln_prior(priors, new_sample)
        # Convert from dictionary of arrays -> array,
        # then filtering out where ln_prior is -infinity
        accepted_samples = np.array(list(new_sample.values()))[:, _ln_prior != -np.inf]
        samples = np.append(samples, accepted_samples, axis=1)
    samples = samples[:, 0:size]
    return samples.transpose()


def sample_best_half(
    param_names: list, priors: dict, wrappedblackbox: callable, size=10
) -> np.ndarray:
    start_points = sample_from_priors(param_names, priors, size=2 * size)
    ln_post = []
    for idx in range(start_points.shape[0]):
        ln_post.append(wrappedblackbox(start_points[idx, :]))
    index_best_half = np.argsort(ln_post)[:size]
    best_points = start_points[index_best_half, :]
    return best_points


def sample_from_high_density_region(
    param_names: list, priors: dict, optimiser, nwalkers: int, nsamples=100
) -> np.ndarray:
    # TODO: remove repeated code
    start_points = sample_from_priors(param_names, priors, size=nsamples)

    ln_prob, _ = optimiser.compute_log_prob(start_points)
    num_best_points = 3
    index_best_start = np.argsort(ln_prob)[-num_best_points:]
    best_start_points = start_points[index_best_start, :]
    best_points_std = np.std(best_start_points, axis=0)

    # Passing samples through ln_prior and redrawing if they fail
    samples = np.empty((param_names.__len__(), 0))
    while samples.size < param_names.__len__() * nwalkers:
        sample = np.random.normal(
            np.mean(best_start_points, axis=0),
            best_points_std,
            size=(nwalkers * 5, len(param_names)),
        )
        start = {name: sample[:, idx] for idx, name in enumerate(param_names)}
        _ln_prior = ln_prior(
            priors,
            start,
        )
        # Convert from dictionary of arrays -> array,
        # then filtering out where ln_prior is -infinity
        accepted_samples = np.array(list(start.values()))[:, _ln_prior != -np.inf]
        samples = np.append(samples, accepted_samples, axis=1)
    start_points = samples[:, 0:nwalkers].transpose()
    return start_points



import matplotlib.pyplot as plt

from indica.profilers.profiler_spline import ProfilerCubicSpline
from indica.profilers.profiler_spline import ProfilerMonoSpline

xknots_mspline = [0.0, 0.15, 0.3, 1.0, 1.05]
xknots_cspline = [
    0.0,
    0.15,
    0.3,
    1.0,
]

_prof = ProfilerMonoSpline(
    datatype="ion_temperature",
    parameters={
        "y0": 1.1,
        "y1": 0.8,
        "y2": 0.6,
        "y3": 0.01,
        "y4": 0.001,
        "xknots": xknots_mspline,
    },
)
_prof.plot(fig=False, color="red", marker="o", label="monospline1")

_prof = ProfilerMonoSpline(
    datatype="ion_temperature",
    parameters={
        "y0": 1,
        "y1": 1.05,
        "y2": 0.8,
        "y3": 0.01,
        "y4": 0.001,
        "xknots": [0.0, 0.2, 0.4, 1.0, 1.05],
    },
)
_prof.plot(fig=False, color="red", marker="x", label="monospline1")

_prof = ProfilerCubicSpline(
    datatype="ion_temperature",
    parameters={
        "y0": 1.1,
        "y1": 0.01,
        "shape1": 0.8,
        "shape2": 0.75,
        "xknots": xknots_cspline,
    },
)
_prof.plot(fig=False, color="blue", marker="o", label="cubicspline1")

_prof = ProfilerCubicSpline(
    datatype="ion_temperature",
    parameters={
        "y0": 1,
        "y1": 0.01,
        "shape1": 1.05,
        "shape2": 0.8,
        "xknots": xknots_cspline,
    },
)
_prof.plot(fig=False, color="blue", marker="x", label="cubicspline1")
plt.legend()
plt.show(block=True)



from pathlib import Path
import pickle


PROJECT_PATH = Path(__file__).parent.parent
DEFAULTS_PATH = f"{PROJECT_PATH}/defaults/"


def get_filename_default_objects(machine: str):
    _files = {}
    _files["geometry"] = (
        DEFAULTS_PATH + f"{machine}_default_geometry_transform_objects.pkl"
    )
    _files["equilibrium"] = DEFAULTS_PATH + f"{machine}_default_equilibrium_object.pkl"
    _files["plasma"] = DEFAULTS_PATH + f"{machine}_default_plasma_object.pkl"

    return _files


def load_default_objects(machine: str, identifier: str = "geometry"):
    """
    Load default objects from local pickle files

    Parameters
    ----------
    machine - e.g. "st40"
    identifier - "geometry" or "equilibrium" or "plasma"
    """
    _file = get_filename_default_objects(machine)[identifier]

    try:
        return pickle.load(open(_file, "rb"))
    except FileNotFoundError:
        to_print = f"""

************************************************************
The following file does not exist:
{_file}

Create your defaults file:
    python indica/defaults/read_write_defaults.py

or, to choose specific kwargs (e.g. machine or pulse):
    save_default_objects("st40", 11419)

************************************************************
            """
        raise Exception(to_print)



import matplotlib.cm as cm
import matplotlib.pylab as plt
import numpy as np
import xarray as xr
from xarray import DataArray

from indica.available_quantities import READER_QUANTITIES
from indica.converters import LineOfSightTransform
from indica.models.abstract_diagnostic import AbstractDiagnostic
from indica.numpy_typing import LabeledArray
import indica.physics as ph
from indica.readers.marchuk import MARCHUKReader
from indica.utilities import build_dataarrays
from indica.utilities import get_element_info
from indica.utilities import set_axis_sci
from indica.utilities import set_plot_rcparams

LINE_RANGES = {
    "w": slice(0.39489, 0.39494),
    "n3": slice(0.39543, 0.39574),
    "n345": slice(0.39496, 0.39574),
    "qra": slice(0.39810, 0.39865),
    "k": slice(0.39890, 0.39910),
    "z": slice(0.39935, 0.39950),
}


class HelikeSpectrometer(AbstractDiagnostic):
    """
    Data and methods to model XRCS spectrometer measurements

    TODO: calibration and Etendue to be correctly included
    TODO: move window to call instead of init
    """

    def __init__(
        self,
        name: str,
        instrument_method="get_helike_spectroscopy",
        etendue: float = 1.0,
        calibration: float = 5.0e-19,
        element: str = "ar",
        window_len: int = 1030,
        window_lim=None,
        window: np.array = None,
        window_masks=None,
        line_labels=None,
        background=0,
        instrumental_broadening: float = 100,
    ):
        """
        Read all atomic data and initialise objects

        Parameters
        ----------
        name
            String identifier for the spectrometer

        """
        self.transform: LineOfSightTransform
        if window_lim is None:
            window_lim = [0.394, 0.401]
        if window_masks is None:
            window_masks = []
        if line_labels is None:
            line_labels = ["w", "k", "n3", "n345", "z", "qra"]

        self.name = name
        self.instrument_method = instrument_method
        self.quantities = READER_QUANTITIES[self.instrument_method]
        self.element: str = element
        z_elem, a_elem, name_elem, _ = get_element_info(element)
        self.ion_charge: int = z_elem - 2  # He-like
        self.ion_mass: float = a_elem
        self.etendue = etendue
        self.calibration = calibration
        self.window_masks = window_masks
        self.line_ranges = LINE_RANGES
        self.line_labels = line_labels
        self.background = background
        self.instrumental_broadening = instrumental_broadening

        if window is None:
            window = np.linspace(window_lim[0], window_lim[1], window_len)
        mask = np.zeros(shape=window.shape)
        if self.window_masks:
            for mslice in self.window_masks:
                mask[(window > mslice[0]) & (window < mslice[1])] = 1
        else:
            mask[:] = 1
        self.window = DataArray(mask, coords={"wavelength": window})
        self._get_atomic_data(self.window)

        self.line_emission: dict
        self.emission_los: dict
        self.measured_intensity: dict
        self.measured_Te: dict
        self.measured_Ti: dict
        self.measured_Nimp: dict
        self.pos: dict
        self.pos_err_in: dict
        self.pos_err_out: dict
        self.spectra: DataArray

        self.Ti: DataArray
        self.Te: DataArray
        self.Ne: DataArray
        self.Nimp: DataArray
        self.Fz: dict
        self.Nh: DataArray

    def _get_atomic_data(self, window: DataArray):
        """
        Read atomic data and keep only lines in desired wavelength window
        """
        marchuck_reader = MARCHUKReader(
            element=self.element,
            charge=self.ion_charge,
        )

        pecs = marchuck_reader.pecs
        mask = window.interp(wavelength=pecs.wavelength)
        pecs = pecs.where(
            mask,
            drop=True,
        )
        pecs = pecs.where(
            (pecs.wavelength < window.wavelength.max())
            & (pecs.wavelength > window.wavelength.min()),
            drop=True,
        )
        self.mask = mask
        self.pecs = pecs

    def _transition_matrix(self, element="ar", charge=16):
        """vectorisation of the transition matrix used to convert
        Helike PECs to emissivity"""
        # fmt: off
        _Nimp = self.Nimp.sel(element=element, )
        _Fz = self.Fz[element]
        transition_matrix = xr.concat([
            self.Ne * _Nimp * _Fz.sel(ion_charge=charge, ),
            self.Ne * _Nimp * _Fz.sel(ion_charge=charge, ),
            self.Ne * _Nimp * _Fz.sel(ion_charge=charge - 1, ),
            self.Ne * _Nimp * _Fz.sel(ion_charge=charge - 1, ),
            self.Ne * _Nimp * _Fz.sel(ion_charge=charge - 1, ),
            self.Ne * _Nimp * _Fz.sel(ion_charge=charge + 1, ),
            self.Nh * _Nimp * _Fz.sel(ion_charge=charge + 1, ),
        ], "type").assign_coords(
            type=["excit", "diel", "li_diel", "ise", "isi", "recom", "cxr", ])
        # fmt: on
        return transition_matrix

    def _calculate_intensity(
        self,
    ):
        """
        Returns DataArrays of emission type with co-ordinates of line label and
        spatial co-ordinate
        """
        mult = self._transition_matrix(element=self.element, charge=self.ion_charge)

        # Swapping to dataset and then dropping line_names with NaNs is much faster
        _pecs_ds = self.pecs.to_dataset("type")
        temp = [
            _pecs_ds[type]
            .dropna("line_name", how="all")
            .interp(electron_temperature=self.Te, assume_sorted=True)
            for type in _pecs_ds.data_vars.keys()
        ]
        _intensity = xr.merge(temp).to_array("type")
        intensity = (_intensity * mult * self.calibration).sum("type")
        self.intensity = intensity

        return intensity

    def _make_spectra(self, calc_rho: bool = False):
        """
        TODO: Doppler Shift / Add convolution of broadening
        -> G(x, mu1, sig1) * G(x, mu2, sig2) = G(x, mu1+mu2, sig1**2 + sig2**2)
        instrument function / photon noise / photon -> counts
        Background Noise

        """
        _spectra = ph.doppler_broaden(
            self.window[self.window > 0].wavelength,
            self.intensity,
            self.intensity.wavelength,
            self.ion_mass,
            self.Ti + self.instrumental_broadening,
        )
        _spectra = _spectra.sum("line_name")
        # extend spectra to same coords as self.window.wavelength with NaNs
        # to maintain same shape as mds data
        if "t" in _spectra.dims:
            empty = xr.DataArray(
                np.nan,
                dims=_spectra.dims,
                coords=dict(
                    wavelength=self.window[self.window < 1].wavelength,
                    rhop=_spectra.rhop,
                    t=_spectra.t,
                ),
            )
        else:
            empty = xr.DataArray(
                np.nan,
                dims=_spectra.dims,
                coords=dict(
                    wavelength=self.window[self.window < 1].wavelength,
                    rhop=_spectra.rhop,
                ),
            )
        spectra = xr.concat([_spectra, empty], "wavelength")
        self.spectra = spectra

        measured_spectra = self.transform.integrate_on_los(
            self.spectra,
            t=self.spectra.t,
            calc_rho=calc_rho,
        )
        measured_spectra = measured_spectra.assign_coords(
            {"wavelength": self.spectra.wavelength}
        )
        measured_spectra = xr.where(measured_spectra == 0, np.nan, measured_spectra)
        self.measured_spectra = measured_spectra.sortby("wavelength")
        self.spectra_los = self.transform.along_los

    def _moment_analysis(self):
        """
        Perform moment analysis using a lines defined in line_labels,
        calculating the position of emissivity, and expected Te and Ti
        """

        line_emission = {}
        for line_name in self.line_labels:
            _line_emission = self.intensity.where(
                (self.intensity.wavelength > self.line_ranges[line_name].start)
                & (self.intensity.wavelength < self.line_ranges[line_name].stop),
                drop=True,
            )
            line_emission[line_name] = _line_emission.sum("line_name")

        if "k" in line_emission.keys() and "w" in line_emission.keys():
            line_emission["kw"] = line_emission["k"] * line_emission["w"]
        if "n3" in line_emission.keys() and "w" in line_emission.keys():
            line_emission["n3w"] = line_emission["n3"] * line_emission["w"]
        if (
            "n3" in line_emission.keys()
            and "n345" in line_emission.keys()
            and "w" in line_emission.keys()
        ):
            line_emission["tot"] = line_emission["n345"] + line_emission["w"]
            line_emission["n3tot"] = line_emission["tot"]
        self.line_emission = line_emission

        rho_mean = {}
        rho_err_in = {}
        rho_err_out = {}
        measured_intensity = {}
        emission_los = {}
        measured_Te = {}
        measured_Ti = {}
        measured_Nimp = {}
        for line in self.line_emission.keys():
            channels = self.transform.x1
            if len(self.transform.x1) == 1:
                channels = self.transform.x1[0]

            emission = self.line_emission[line]
            los_integral = self.transform.integrate_on_los(emission, t=emission.t)
            emission_los = self.transform.along_los.sel(channel=channels).mean(
                "beamlet"
            )
            emission_sum = emission_los.sum("los_position", skipna=True)
            rho_los = self.transform.rhop.sel(channel=channels)

            rho_mean[line] = (emission_los * rho_los).sum(
                "los_position", skipna=True
            ) / emission_sum
            rho_err = rho_los  # rho_los.where(indx_err, np.nan,)
            where_in = rho_err < rho_mean[line]
            where_out = rho_err > rho_mean[line]
            rho_in = xr.where(where_in, rho_los, np.nan)
            rho_out = xr.where(where_out, rho_los, np.nan)
            rho_err_in[line] = (
                (emission_los * (rho_in - rho_mean[line]) ** 2).sum(
                    "los_position", skipna=True
                )
                / emission_sum
            ) ** 0.5
            rho_err_out[line] = (
                (emission_los * (rho_out - rho_mean[line]) ** 2).sum(
                    "los_position", skipna=True
                )
                / emission_sum
            ) ** 0.5
            measured_intensity[line] = los_integral
            emission_los[line] = emission_los

            Te_along_los = self.transform.map_profile_to_los(self.Te, t=emission.t).sel(
                channel=channels
            )
            measured_Te[line] = (emission_los * Te_along_los).sum(
                "los_position", skipna=True
            ) / emission_sum

            Ti_along_los = self.transform.map_profile_to_los(self.Ti, t=emission.t).sel(
                channel=channels
            )
            measured_Ti[line] = (emission_los * Ti_along_los).sum(
                "los_position", skipna=True
            ) / emission_sum

            Nimp_along_los = self.transform.map_profile_to_los(
                self.Nimp.sel(element=self.element), t=emission.t
            ).sel(channel=channels)
            measured_Nimp[line] = (emission_los * Nimp_along_los).sum(
                "los_position", skipna=True
            ) / emission_sum

        self.pos = rho_mean
        self.pos_err_in = rho_err_in
        self.pos_err_out = rho_err_out
        self.measured_intensity = measured_intensity
        self.emission_los = emission_los
        self.measured_Te = measured_Te
        self.measured_Ti = measured_Ti
        self.measured_Nimp = measured_Nimp

    def _build_bckc_dictionary(self):
        bckc = {
            "t": self.t,
            "channel": np.arange(len(self.transform.x1)),
            "wavelength": self.window.wavelength,
            "location": self.transform.origin,
            "direction": self.transform.direction,
            "spectra_raw": self.measured_spectra,
        }

        if self.moment_analysis:
            for line in self.measured_intensity.keys():
                self.bckc[f"te_{line}"] = self.measured_Te[line]
                self.bckc[f"ti_{line}"] = self.measured_Ti[line]
                self.bckc[f"int_{line}"] = self.measured_intensity[line]
                self.bckc[f"int_{line}"].attrs["emiss"] = self.line_emission[line]
                self.bckc[f"int_{line}"].attrs["pos"] = self.pos[line]
                self.bckc[f"int_{line}"].attrs["pos_err_in"] = self.pos_err_in[line]
                self.bckc[f"int_{line}"].attrs["pos_err_out"] = self.pos_err_out[line]
            self.bckc["int_k/int_w"] = self.bckc["int_k"] / self.bckc["int_w"]
            self.bckc["int_n3/int_w"] = self.bckc["int_n3"] / self.bckc["int_w"]
            self.bckc["int_n3/int_tot"] = self.bckc["int_n3"] / self.bckc["int_tot"]

        self.bckc = build_dataarrays(bckc, self.quantities, transform=self.transform)

    def __call__(
        self,
        Te: DataArray = None,
        Ti: DataArray = None,
        Ne: DataArray = None,
        Nimp: DataArray = None,
        Fz: dict = None,
        Nh: DataArray = None,
        t: LabeledArray = None,
        calc_rho: bool = False,
        moment_analysis: bool = False,
        background: float = None,
        pixel_offset: int = None,
        norm_spectra: xr.DataArray = None,
        scale_spectra: float = 1.0,
        **kwargs,
    ):
        """
        Calculate diagnostic measured values

        Parameters
        ----------
        Te - electron temperature (eV)
        Ti - ion temperature (eV)
        Ne - electron density (m**-3)
        Nimp - impurity density (m**-3)
        fractional_abundance - fractional abundance
        Nh - neutral density (m**-3)
        t - time (s)

        Returns
        -------

        """
        self.calc_rho = calc_rho
        self.moment_analysis = moment_analysis

        if self.moment_analysis and bool(self.window_masks):
            raise ValueError(
                "moment_analysis cannot be used when window_masks is not set to None"
            )

        if hasattr(self, "plasma"):
            if t is None:
                t = self.plasma.time_to_calculate
            Te = self.plasma.electron_temperature.sel(
                t=t,
            )
            Ne = self.plasma.electron_density.sel(
                t=t,
            )
            Nh = self.plasma.neutral_density.sel(
                t=t,
            )
            Fz = {}
            _Fz = self.plasma.fz
            for elem in _Fz.keys():
                Fz[elem] = _Fz[elem].sel(t=t)

            Ti = self.plasma.ion_temperature.sel(t=t)
            Nimp = self.plasma.impurity_density.sel(t=t)
        else:
            if (
                Ne is None
                or Te is None
                or Nh is None
                or Fz is None
                or Ti is None
                or Nimp is None
            ):
                raise ValueError("Give inputs or assign plasma class!")

        if background is None:
            if hasattr(self.background, "sel"):
                background = self.background.sel(t=t)
            else:
                background = self.background

        self.t = t
        self.Te = Te
        self.Ne = Ne
        self.Nh = Nh
        self.Fz = Fz
        self.Ti = Ti
        self.Nimp = Nimp

        # TODO: check that inputs have compatible dimensions/coordinates
        self._calculate_intensity()
        self._make_spectra()

        if moment_analysis:
            self._moment_analysis()

        dt = np.gradient(self.plasma.t).mean()  # Temp hack for count -> count / s
        self.measured_spectra = self.measured_spectra / dt

        if norm_spectra is not None:
            self.measured_spectra = (
                self.measured_spectra
                / self.measured_spectra.max()
                * (norm_spectra.sel(t=t) * scale_spectra - background)
            )

        self.measured_spectra += background

        if pixel_offset is not None:
            self.measured_spectra = self.measured_spectra.shift(
                wavelength=round(pixel_offset), fill_value=np.nan
            )

        self._build_bckc_dictionary()
        return self.bckc

    def plot(self):
        set_plot_rcparams("profiles")

        self.transform.plot(np.mean(self.t))

        plt.figure()
        channels = self.transform.x1
        cols_time = cm.gnuplot2(np.linspace(0.1, 0.75, len(self.t), dtype=float))
        if "spectra_raw" in self.bckc.keys():
            spectra_raw = self.bckc["spectra_raw"]
            if "channel" in spectra_raw.dims:
                spectra_raw = spectra_raw.sel(channel=np.median(channels))
            for i, t in enumerate(np.array(self.t, ndmin=1)):
                plt.plot(
                    spectra_raw.wavelength,
                    spectra_raw.sel(t=t),
                    color=cols_time[i],
                    label=f"t={t:1.2f} s",
                )
            plt.ylabel("Intensity (count/s.)")
            plt.xlabel("Wavelength (nm)")
            plt.legend()
            set_axis_sci()

        # Plot the temperatures profiles
        plt.figure()
        for i, t in enumerate(self.t):
            plt.plot(
                self.plasma.ion_temperature.rhop,
                self.plasma.ion_temperature.sel(
                    t=t,
                ),
                color=cols_time[i],
            )
            plt.plot(
                self.plasma.electron_temperature.rhop,
                self.plasma.electron_temperature.sel(t=t),
                color=cols_time[i],
                linestyle="dashed",
            )
        plt.xlabel("rho")
        set_axis_sci()
        plt.ylabel("Ti and Te profiles (eV)")

        # Plot the emission profiles
        if self.moment_analysis:
            if "w" in self.line_emission.keys():
                plt.figure()
                if "t" in self.line_emission["w"].dims:
                    for i, t in enumerate(self.plasma.t.values):
                        plt.plot(
                            self.line_emission["w"].rhop,
                            self.line_emission["w"].sel(t=t),
                            color=cols_time[i],
                            label=f"t={t:1.2f} s",
                        )
                else:
                    plt.plot(
                        self.line_emission["w"].rhop,
                        self.line_emission["w"],
                        color=cols_time[i],
                        label=f"t={t:1.2f} s",
                    )
                plt.xlabel("rho")
                plt.ylabel("w-line emissivity (W/m^3)")
                set_axis_sci()
                plt.legend()

        # Plot moment analysis of measured temperatures
        if "ti_w" in self.bckc.keys() & "te_kw" in self.bckc.keys():
            plt.figure()
            for i, t in enumerate(self.plasma.t.values):
                self.bckc["ti_w"].mean("beamlet").sel(t=t).plot(
                    label=f"t={t:1.2f} s",
                    marker="o",
                    color=cols_time[i],
                )
                self.bckc["te_kw"].mean("beamlet").sel(t=t).plot(
                    color=cols_time[i],
                    marker="x",
                    linestyle="dashed",
                )

            plt.xlabel("Channel")
            plt.ylabel("Measured Ti and Te (eV)")
            set_axis_sci()
            plt.title("")
            # plt.legend()

        plt.show(block=True)



"""Trivial class for transforming to and from R-z coordinate systems.
"""

from .abstractconverter import Coordinates
from .abstractconverter import CoordinateTransform
from ..numpy_typing import LabeledArray


class TrivialTransform(CoordinateTransform):
    """Class to use for transforms to/from an R-z coordinate systems. This
    is trivial, as R-z coordinates are the go-between for conversion
    to all other coordinate systems.

    """

    x1_name = "R"
    x2_name = "z"

    def convert_to_Rz(
        self, x1: LabeledArray, x2: LabeledArray, t: LabeledArray
    ) -> Coordinates:
        """Convert from this coordinate to the R-z coordinate system.

        Parameters
        ----------
        x1
            The first spatial coordinate in this system.
        x2
            The second spatial coordinate in this system.
        t
            The time coordinate (if there is one, otherwise ``None``)

        Returns
        -------
        R
            Major radius coordinate
        z
            Height coordinate

        """
        return x1, x2

    def convert_from_Rz(
        self, R: LabeledArray, z: LabeledArray, t: LabeledArray
    ) -> Coordinates:
        """Convert from the master coordinate system to this coordinate.

        Parameters
        ----------
        R
            Major radius coordinate
        z
            Height coordinate
        t
            Time coordinate)

        Returns
        -------
        x1
            The first spatial coordinate in this system.
        x2
            The second spatial coordinate in this system.

        """
        return R, z

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        return self._abstract_equals(other)



from typing import Tuple

import numpy as np

from indica import Plasma
from indica import PlasmaProfiler
from indica.profilers.profiler_gauss import initialise_gauss_profilers


def example_plasma(
    tstart=0.02,
    tend=0.1,
    dt=0.01,
    main_ion="h",
    impurities: Tuple[str, ...] = ("c", "ar", "he"),
    **kwargs,
):
    plasma = Plasma(
        tstart=tstart,
        tend=tend,
        dt=dt,
        main_ion=main_ion,
        impurities=impurities,
        **kwargs,
    )
    plasma.build_atomic_data()

    profilers = initialise_gauss_profilers(
        plasma.rhop,
        profile_names=[
            "electron_density",
            "ion_temperature",
            "toroidal_rotation",
            "electron_temperature",
            "impurity_density:ar",
            "impurity_density:c",
            "impurity_density:he",
        ],
    )
    plasma_profiler = PlasmaProfiler(
        plasma=plasma,
        profilers=profilers,
    )
    plasma_profiler()

    # Assign profiles to time-points
    nt = len(plasma.t)
    ne_peaking = np.linspace(1, 2, nt)
    te_peaking = np.linspace(1, 2, nt)
    _y0 = plasma_profiler.profilers["toroidal_rotation"].y0
    vrot0 = np.linspace(
        _y0 * 1.1,
        _y0 * 2.5,
        nt,
    )
    vrot_peaking = np.linspace(1, 2, nt)

    _y0 = plasma_profiler.profilers["ion_temperature"].y0
    ti0 = np.linspace(_y0 * 1.1, _y0 * 2.5, nt)

    _y0 = plasma_profiler.profilers[f"impurity_density:{impurities[0]}"].y0
    nimp_y0 = _y0 * 5 * np.linspace(1, 8, nt)
    nimp_peaking = np.linspace(1, 5, nt)
    nimp_wcenter = np.linspace(0.4, 0.1, nt)
    for i, t in enumerate(plasma.t):
        parameters = {
            "electron_temperature.peaking": te_peaking[i],
            "ion_temperature.peaking": te_peaking[i],
            "ion_temperature.y0": ti0[i],
            "toroidal_rotation.peaking": vrot_peaking[i],
            "toroidal_rotation.y0": vrot0[i],
            "electron_density.peaking": ne_peaking[i],
            "impurity_density:ar.peaking": nimp_peaking[i],
            "impurity_density:ar.y0": nimp_y0[i],
            "impurity_density:ar.wcenter": nimp_wcenter[i],
        }

        plasma_profiler(parameters=parameters, t=t)

    return plasma



"""Provides an abstract interface for coordinate conversion.
"""
from abc import ABC
from abc import abstractmethod
import itertools
from typing import Callable
from typing import cast
from typing import Dict
from typing import Optional
from typing import Tuple

from matplotlib import cm
from matplotlib import rcParams
import matplotlib.pylab as plt
import numpy as np
import xarray as xr
from xarray import DataArray
from xarray import zeros_like

from indica.utilities import FIG_PATH
from indica.utilities import intersection
from indica.utilities import save_figure
from indica.utilities import set_plot_rcparams
from ..equilibrium import Equilibrium
from ..numpy_typing import ArrayLike
from ..numpy_typing import Coordinates
from ..numpy_typing import LabeledArray
from ..numpy_typing import OnlyArray


class EquilibriumException(Exception):
    """Exception raised if a converter object's equilibrium object is set
    twice."""


class CoordinateTransform(ABC):
    """Class for converting between different coordinate systems. This is
    an abstract base class; each coordinate system should provide its own
    implementation.

    Subclasses should allow each instance to have a "default grid" on
    which to calculate results. This can be cached for efficient
    retrieval.

    Note that not all coordinate systems will have an actual x2
    dimension (for example, the lines-of-site for soft X-ray
    data). However, 2 coordinates are still needed to map to the
    global coordinate system. Therefore, x2 is treated as a
    "pseudo-coordinate",in these cases, with values between 0 and 1
    specifying the position along the grid-line for x1. 0 is the start
    and 1 is the end (possibly overlapping, if the coordinate system is
    periodic).

    Parameters
    ----------
    default_x1
        The default grid to use for the first spatial coordinate.
    default_x1
        The default grid to use for the second spatial coordinate.
    default_R
        The default grid to use for the R-coordinate when converting to this
        coordinate system.
    default_z
        The default grid to use for the z-coordinate when converting to this
        coordinate system.
    default_t
        The default grid to use for time.

    Attributes
    ----------
    x1_name: str
        Name for the first spacial coordinate. May be class- or
        instance-specific.

    x2_name: str
        Name for the second spacial coordinate. May be class- or
        instance-specific.
    """

    _CONVERSION_METHODS: Dict[str, str] = {}
    _INVERSE_CONVERSION_METHODS: Dict[str, str] = {}

    equilibrium: Equilibrium
    x1_name: str
    x2_name: str
    x1: LabeledArray
    x2: LabeledArray
    t: LabeledArray = None
    name: str
    instrument_name: str
    _machine_dims: Tuple[Tuple[float, float], Tuple[float, float]]

    dl: float
    x: DataArray
    y: DataArray
    z: DataArray
    R: DataArray
    phi: DataArray
    rhop: DataArray
    theta: DataArray
    profile_to_map: DataArray
    along_los: DataArray
    los_integral: DataArray
    impact_parameter: DataArray
    _origin: OnlyArray
    _direction: OnlyArray

    def set_equilibrium(self, equilibrium: Equilibrium, force: bool = False):
        """Initialise the object using a set of equilibrium data.

        If it has already been initialised with the same equilibrium
        data then do nothing. If already initialised with a different
        equilibrium, throw an
        :py:class:`abstractconverter.EquilibriumException` unless
        ``force == True``.

        Parameters
        ----------
        equilibrium
            A set of equilibrium data with which to calculate coordinate
            transforms.
        force : bool
            If true, re-initialise the transform if provided with a new set of
            equilibrium data.

        """
        if not hasattr(self, "equilibrium") or force:
            self.equilibrium = equilibrium
            if hasattr(self, "rhop"):
                delattr(self, "rhop")
        elif self.equilibrium != equilibrium:
            raise EquilibriumException("Attempt to set equilibrium twice.")

    def check_equilibrium(self):
        if not hasattr(self, "equilibrium"):
            raise Exception("Missing equilibrium object")

    def get_converter(
        self, other: "CoordinateTransform", reverse=False
    ) -> Optional[Callable[[LabeledArray, LabeledArray, LabeledArray], Coordinates]]:
        """Checks if there is a shortcut to convert between these coordiantes,
        returning it if so. This can sometimes save the step of
        converting to (R, z) coordinates first.

        Parameters
        ----------
        other
            The other transform whose coordinate system you want to convert to.
        reverse
            If True, try to return a function which converts _from_ ``other``
            to this coordinate system.

        Returns
        -------
        :
            If a shortcut function is available, return it. Otherwise, None.

        Note
        ----
        Implementations should call ``other.get_converter(self, reverse=True``. For
        obvious reasons, however, they should **only do this when
        ``reverse == False``**.

        """
        if reverse:
            return None
        return other.get_converter(self, True)

    def convert_to(
        self,
        other: "CoordinateTransform",
        x1: LabeledArray,
        x2: LabeledArray,
        t: LabeledArray,
    ) -> Coordinates:
        """General routine to map coordinates from this system to those used
        in ``other``. Array broadcasting will be performed as necessary.

        If this transform class provides a specialised method for
        doing this (specified in :py:attr:`_CONVERSION_METHODS`) then that is
        used. Otherwise, the coordinates are converted to R-z using
        :py:meth:`_convert_to_Rz` and then converted to the other coordinate
        system using :py:attr:`_convert_from_Rz`.

        Parameters
        ----------
        other
            The coordinate system to convert to.
        x1
            The first spatial coordinate in this system.
        x2
            The second spatial coordinate in this system.
        t
            The time coordinate (if there is one, otherwise ``None``)

        Returns
        -------
        x1
            The first spatial coordinate in the ``other`` system.
        x2
            The second spatial coordinate in the ``other`` system.

        """
        if self == other:
            return x1, x2
        converter = self.get_converter(other)
        if converter:
            return converter(x1, x2, t)
        R, z = self.convert_to_Rz(x1, x2, t)
        return other.convert_from_Rz(R, z, t)

    @abstractmethod
    def convert_to_Rz(
        self,
        x1: LabeledArray,
        x2: LabeledArray,
        t: LabeledArray,
    ) -> Coordinates:
        """Convert from this coordinate to the R-z coordinate system. Each
        subclass must implement this method.

        Parameters
        ----------
        x1
            The first spatial coordinate in this system.
        x2
            The second spatial coordinate in this system.
        t
            The time coordinate

        Returns
        -------
        R
            Major radius coordinate
        z
            Height coordinate

        """
        raise NotImplementedError(
            "{} does not implement a 'convert_to_Rz' "
            "method.".format(self.__class__.__name__)
        )

    def convert_to_xy(
        self,
        x1: LabeledArray,
        x2: LabeledArray,
        t: LabeledArray,
    ) -> Coordinates:
        """Convert from this coordinate to the x-y coordinate system. Each
        subclass must implement this method.

        Parameters
        ----------
        x1
            The first spatial coordinate in this system.
        x2
            The second spatial coordinate in this system.
        t
            The time coordinate

        Returns
        -------
        R
            Major radius coordinate
        z
            Height coordinate

        """
        raise NotImplementedError(
            "{} does not implement a 'convert_to_xy' "
            "method.".format(self.__class__.__name__)
        )

    def convert_from_Rz(
        self,
        R: LabeledArray,
        z: LabeledArray,
        t: LabeledArray,
    ) -> Coordinates:
        """Convert from the master coordinate system to this coordinate. Each
        subclass must implement this method.

        Parameters
        ----------
        R
            Major radius coordinate
        z
            Height coordinate
        t
            Time coordinate

        Returns
        -------
        x1
            The first spatial coordinate in this system.
        x2
            The second spatial coordinate in this system.

        """
        raise NotImplementedError(
            "{} does not implement a 'convert_from_Rz' "
            "method.".format(self.__class__.__name__)
        )

    def _abstract_equals(self, other: "CoordinateTransform") -> bool:
        """Checks that default coordinate values and equilibrium objects are
        the same on two transform classes.

        """
        if not hasattr(self, "equilibrium"):
            return not hasattr(other, "equilibrium")
        elif not hasattr(other, "equilibrium"):
            return False
        else:
            result = self.equilibrium == other.equilibrium
            result = result and self.x1_name == other.x1_name
            result = result and self.x2_name == other.x2_name
            return result

    @abstractmethod
    def __eq__(self, other: object) -> bool:
        """Check that two transforms are describing the same coordinate system."""
        raise NotImplementedError(
            "{} does not implement an '__eq__' method".format(self.__class__.__name__)
        )

    def distance(
        self,
        direction: str,
        x1: LabeledArray,
        x2: LabeledArray,
        t: LabeledArray,
    ) -> LabeledArray:
        """Give the distance (in physical space) from the origin along the
        specified direction.

        This is useful for when taking spatial integrals and differentials in
        that direction.

        Note that distance is calculated using Euclidean lines between
        points. As such, it will not be accurate for a curved axis.

        Parameters
        ----------
        direction : str
            Which dimension to give the distance along.
        x1
            The first spatial coordinate in this system.
        x2
            The second spatial coordinate in this system.
        t
            The time coordinate

        Returns
        -------
        :
           Distance from the origin in the specified direction.

        """
        R, z = cast(Tuple[DataArray, DataArray], self.convert_to_Rz(x1, x2, t))
        if isinstance(R, (int, float)) or isinstance(z, (int, float)):
            raise ValueError("Arguments x1 and x2 must be xarray DataArray objects.")
        spacings = np.sqrt(R.diff(direction) ** 2 + z.diff(direction) ** 2)
        result = zeros_like(R.broadcast_like(z))
        result[{direction: slice(1, None)}] = spacings.cumsum(direction)
        return result

    def encode(self) -> str:
        """Returns a JSON representation of this object. Should be sufficient
        to recreate it identically from scratch (except for the
        equilibrium)."""
        return ""

    @staticmethod
    def decode(json: str) -> "CoordinateTransform":
        """Takes some JSON and decodes it into a CoordinateTransform object."""
        pass

    def get_machine_boundaries(
        self,
        machine_dimensions: Tuple[Tuple[float, float], Tuple[float, float]] = (
            (1.83, 3.9),
            (-1.75, 2.0),
        ),
        npts: int = 1000,
    ) -> Tuple[dict, ArrayLike]:
        angles = np.linspace(0.0, 2 * np.pi, npts)
        x_wall_inner = machine_dimensions[0][0] * np.cos(angles)
        x_wall_outer = machine_dimensions[0][1] * np.cos(angles)
        y_wall_inner = machine_dimensions[0][0] * np.sin(angles)
        y_wall_outer = machine_dimensions[0][1] * np.sin(angles)
        z_wall_lower = machine_dimensions[1][0]
        z_wall_upper = machine_dimensions[1][1]

        boundaries = {
            "x_in": x_wall_inner,
            "x_out": x_wall_outer,
            "y_in": y_wall_inner,
            "y_out": y_wall_outer,
            "z_up": z_wall_upper,
            "z_low": z_wall_lower,
        }

        return boundaries, angles

    def get_equilibrium_boundaries(
        self, t: float, npts: int = 1000
    ) -> Tuple[dict, ArrayLike, DataArray]:

        boundaries = {}
        angles = np.linspace(0.0, 2 * np.pi, npts)
        if hasattr(self, "equilibrium"):
            angles = np.linspace(0.0, 2 * np.pi, npts)
            rhop_equil = self.equilibrium.rhop.sel(t=t, method="nearest")
            R_hfs = self.equilibrium.rmji.sel(t=t, rhop=1, method="nearest").values
            R_lfs = self.equilibrium.rmjo.sel(t=t, rhop=1, method="nearest").values
            x_plasma_inner = R_hfs * np.cos(angles)
            x_plasma_outer = R_lfs * np.cos(angles)
            y_plasma_inner = R_hfs * np.sin(angles)
            y_plasma_outer = R_lfs * np.sin(angles)

            boundaries = {
                "x_in": x_plasma_inner,
                "x_out": x_plasma_outer,
                "y_in": y_plasma_inner,
                "y_out": y_plasma_outer,
            }
        return boundaries, angles, rhop_equil

    def convert_to_rho_theta(self, t: LabeledArray = None) -> Coordinates:
        """
        Convert R, z to rho, theta given the flux surface transform
        """
        if not hasattr(self, "equilibrium"):
            raise Exception("Set equilibrium object to convert (R,z) to rho")

        rhop, theta, _ = self.equilibrium.flux_coords(self.R, self.z, t=t)
        drop_vars = ["R", "z"]
        for var in drop_vars:
            if var in rhop.coords:
                rhop = rhop.drop_vars(var)
            if var in theta.coords:
                theta = theta.drop_vars(var)

        self.t = t
        self.rhop = rhop
        self.theta = theta
        if "los_position" in self.rhop.dims:
            rhop_mean = self.rhop.mean("beamlet")
            self.impact_rho = rhop_mean.sel(
                los_position=self.impact_parameter.index.los_position
            )
            self.los_length = (xr.where(np.isfinite(rhop_mean), 1, 0) * self.dl).sum(
                "los_position"
            )

        return rhop, theta

    def plot(
        self,
        t: float = None,
        orientation: str = "all",
        figure: bool = True,
        save_fig: bool = False,
        fig_path: str = "",
        fig_name: str = "",
        markersize: float = None,
        marker: str = "o",
    ):

        cols = cm.gnuplot2(
            np.linspace(0.1, 0.75, np.size(np.array(self.x1)), dtype=float)
        )

        if len(fig_path) == 0:
            fig_path = FIG_PATH

        set_plot_rcparams("profiles")
        if markersize is not None:
            rcParams.update({"lines.markersize": markersize})

        wall_bounds, angles = self.get_machine_boundaries(
            machine_dimensions=self._machine_dims
        )

        if hasattr(self, "equilibrium"):
            if t is None:
                t = np.float64(np.mean(self.equilibrium.rhop.t))
            equil_bounds, angles, rhop_equil = self.get_equilibrium_boundaries(t)
            x_ax = self.equilibrium.rmag.sel(t=t, method="nearest").values * np.cos(
                angles
            )
            y_ax = self.equilibrium.rmag.sel(t=t, method="nearest").values * np.sin(
                angles
            )

        title = f"{self.instrument_name.upper()}"
        if t is not None:
            title += f" @ {t:.3f} s"

        trans_name = str(self)

        if orientation == "xy" or orientation == "all":
            if figure:
                plt.figure()
            plt.plot(wall_bounds["x_in"], wall_bounds["y_in"], color="k")
            plt.plot(wall_bounds["x_out"], wall_bounds["y_out"], color="k")
            if hasattr(self, "equilibrium"):
                plt.plot(equil_bounds["x_in"], equil_bounds["y_in"], color="red")
                plt.plot(equil_bounds["x_out"], equil_bounds["y_out"], color="red")
                plt.plot(x_ax, y_ax, color="red", linestyle="dashed")
            plot_geometry(
                self.x,
                self.y,
                trans_name,
                colors=cols,
                marker=marker,
            )
            plt.xlabel("x [m]")
            plt.ylabel("y [m]")
            plt.axis("scaled")
            plt.title(title)
            save_figure(fig_path, f"{fig_name}{self.name}_xy", save_fig=save_fig)

        if orientation == "Rz" or orientation == "all":
            if figure:
                plt.figure()
            plt.plot(
                [wall_bounds["x_out"].max()] * 2,
                [wall_bounds["z_low"], wall_bounds["z_up"]],
                color="k",
            )
            plt.plot(
                [wall_bounds["x_in"].max()] * 2,
                [wall_bounds["z_low"], wall_bounds["z_up"]],
                color="k",
            )
            plt.plot(
                [wall_bounds["x_in"].max(), wall_bounds["x_out"].max()],
                [wall_bounds["z_low"]] * 2,
                color="k",
            )
            plt.plot(
                [wall_bounds["x_in"].max(), wall_bounds["x_out"].max()],
                [wall_bounds["z_up"]] * 2,
                color="k",
            )
            if hasattr(self, "equilibrium"):
                rhop_equil.plot.contour(levels=[0.01, 0.1, 0.3, 0.5, 0.7, 0.9, 0.99])

            plot_geometry(
                self.R,
                self.z,
                trans_name,
                colors=cols,
                marker=marker,
            )
            plt.xlabel("R [m]")
            plt.ylabel("z [m]")
            plt.title(title)
            plt.axis("scaled")
            save_figure(fig_path, f"{fig_name}{self.name}_Rz", save_fig=save_fig)

        if hasattr(self, "equilibrium") and orientation == "all":
            if not hasattr(self, "rhop"):
                self.convert_to_rho_theta(t=[t])
            if figure:
                plt.figure()
            _rhop = self.rhop
            if "t" in self.rhop.dims:
                _rhop = _rhop.sel(t=t, method="nearest")

            if "LineOfSight" in trans_name:
                abscissa = _rhop.los_position.expand_dims(
                    dim={"channel": _rhop.channel, "beamlet": _rhop.beamlet}
                )
            elif "Transect" in trans_name:
                abscissa = _rhop.channel
            plot_geometry(
                abscissa,
                _rhop,
                trans_name,
                colors=cols,
                marker=marker,
            )
            plt.xlabel("Path along LOS")
            plt.ylabel("Rhop")
            plt.title(title)
            save_figure(fig_path, f"{fig_name}{self.name}_rhop", save_fig=save_fig)

        return cols


def plot_geometry(
    abscissa: DataArray,
    ordinate: DataArray,
    trans_name: str,
    colors: ArrayLike,
    marker: str = "o",
):

    if "LineOfSight" in trans_name:
        marker = None
    if hasattr(abscissa, "beamlet"):
        beamlets = abscissa.beamlet.values
    else:
        beamlets = [
            0,
        ]
    channels = abscissa.channel.values
    for channel, beamlet in itertools.product(channels, beamlets):
        col = colors[channel]
        x = abscissa.sel(channel=channel)
        y = ordinate.sel(channel=channel)
        if hasattr(abscissa, "beamlet"):
            x = x.sel(beamlet=beamlet)
            y = y.sel(beamlet=beamlet)
        plt.plot(x, y, color=col, marker=marker)


def find_wall_intersections(
    origin: Tuple,
    direction: Tuple,
    machine_dimensions: Tuple[Tuple[float, float], Tuple[float, float]] = (
        (1.83, 3.9),
        (-1.75, 2.0),
    ),
):
    """Function for calculating "start" and "end" positions of the line-of-sight
    given the machine dimensions.

    The end coordinate is calculated by finding the intersections with the
    machine dimensions. If the intersection hits the inner column, the first
    intersection point becomes the "end" position of the line-of-sight. If the
    intersection misses the inner column, the last intersection point with the
    outer column becomes the "end" position.

    Parameters
    ----------
    origin
        A Tuple (1x3) giving the X, Y and Z origin positions of the line-of-sight
    direction
        A Tuple (1x3) giving the X, Y and Z direction of the line-of-sight
    machine_dimensions
        A tuple giving the boundaries of the Tokamak in x-z space:
        ``((xmin, xmax), (zmin, zmax)``. Defaults to values for JET.

    Returns
    -------
    start_coordinates
        A Tuple (1x3) giving the X, Y and Z start positions of the line-of-sight
    end_coordinates
        A Tuple (1x3) giving the X, Y and Z end positions of the line-of-sight
    """

    def line(_start, _end):
        return np.linspace(_start, _end, npts)

    def extremes(_start, _end):
        return np.array(
            [_start, _end],
            dtype=float,
        )

    # Define XYZ lines for inner and outer walls
    npts = 1000
    angles = np.linspace(0.0, 2 * np.pi, npts)
    x_wall_inner = machine_dimensions[0][0] * np.cos(angles)
    y_wall_inner = machine_dimensions[0][0] * np.sin(angles)

    # Define XYZ lines for LOS from origin and direction vectors
    length = (
        np.ceil(
            np.max(
                [
                    machine_dimensions[0][1] * 2,
                    machine_dimensions[1][1] - machine_dimensions[1][0],
                ]
            )
        )
        * 5
    )
    x_start = origin[0]
    x_end = origin[0] + length * direction[0]
    y_start = origin[1]
    y_end = origin[1] + length * direction[1]
    z_start = origin[2]
    z_end = origin[2] + length * direction[2]

    # Find intersections in R, z plane
    x_line = line(x_start, x_end)
    y_line = line(y_start, y_end)
    z_line = line(z_start, z_end)
    R_line = np.sqrt(x_line**2 + y_line**2)
    indices = np.where(
        (R_line >= machine_dimensions[0][0])
        * (R_line <= machine_dimensions[0][1])
        * (z_line >= machine_dimensions[1][0])
        * (z_line <= machine_dimensions[1][1])
    )[0]
    if len(indices) > 0:
        x_start = x_line[indices][0]
        x_end = x_line[indices][-1]
        y_start = y_line[indices][0]
        y_end = y_line[indices][-1]
        z_start = z_line[indices][0]
        z_end = z_line[indices][-1]

    # Find intersections with inner wall
    xx, yx, ix, jx = intersection(
        extremes(x_start, x_end), extremes(y_start, y_end), x_wall_inner, y_wall_inner
    )
    if len(xx) > 0:
        x_line = line(x_start, x_end)
        y_line = line(y_start, y_end)
        z_line = line(z_start, z_end)
        index = []
        for i in range(len(xx)):
            index.append(
                np.argmin(np.sqrt((xx[i] - x_line) ** 2 + (yx[i] - y_line) ** 2))
            )
        indices = np.arange(np.min(index))
        x_start = x_line[indices][0]
        x_end = x_line[indices][-1]
        y_start = y_line[indices][0]
        y_end = y_line[indices][-1]
        z_start = z_line[indices][0]
        z_end = z_line[indices][-1]

    return (x_start, y_start, z_start), (x_end, y_end, z_end)



import numpy as np


class ST40ReaderProcessorConf:
    def __init__(self):

        self.filter_values = {
            "cxff_pi": {"ti": (0, np.inf), "vtor": (0, np.inf)},
            "cxff_tws_c": {"ti": (0, np.inf), "vtor": (0, np.inf)},
            "cxff_tws_b": {"ti": (0, np.inf), "vtor": (0, np.inf)},
            "cxqf_tws_c": {"ti": (0, np.inf), "vtor": (0, np.inf)},
            "xrcs": {
                "ti_w": (0, np.inf),
                "ti_z": (0, np.inf),
                "te_kw": (0, np.inf),
                "te_n3w": (0, np.inf),
                "spectra": (0, np.inf),
                "spectra_raw": (0, np.inf),
            },
            "brems": {"brightness": (0, np.inf)},
            "halpha": {"brightness": (0, np.inf)},
            "sxr_spd": {"brightness": (0, np.inf)},
            "sxr_diode_1": {"brightness": (0, np.inf)},
            "sxr_camera_4": {"brightness": (0, np.inf)},
            "sxrc_xy1": {"brightness": (0, np.inf)},
            "sxrc_xy2": {"brightness": (0, np.inf)},
            "blom_xy1": {"brightness": (0, np.inf)},
            "ts": {"te": (0, np.inf), "ne": (0, np.inf)},
            "ppts": {
                "te_rho": (1, np.inf),
                "ne_rho": (1, np.inf),
                "pe_rho": (1, np.inf),
                "te_R": (1, np.inf),
                "ne_R": (1, np.inf),
                "pe_R": (1, np.inf),
            },
            "pi": {"spectra": (0, np.inf)},
            "tws_c": {"spectra": (0, np.inf)},
        }

        self.filter_coordinates = {
            "cxff_pi": {
                "ti": ("channel", (0, np.inf)),
                "vtor": ("channel", (0, np.inf)),
            },
            "cxff_tws_c": {
                "ti": ("channel", (0, np.inf)),
                "vtor": ("channel", (0, np.inf)),
            },
            "cxff_tws_b": {
                "ti": ("channel", (0, np.inf)),
                "vtor": ("channel", (0, np.inf)),
            },
            "xrcs": {
                "spectra": ("wavelength", (0.0, np.inf)),
                "spectra_raw": ("wavelength", (0.0, np.inf)),
            },
            "ts": {"te": ("channel", (0, np.inf)), "ne": ("channel", (0, np.inf))},
        }



"""Coordinate system representing a collection of lines of sight.
"""

from typing import Any
from typing import cast
from typing import Tuple

import matplotlib.pylab as plt
import numpy as np
import xarray as xr
from xarray import DataArray
from xarray import Dataset
from xarray import zeros_like

from .abstractconverter import Coordinates
from .abstractconverter import CoordinateTransform
from .abstractconverter import find_wall_intersections
from ..numpy_typing import LabeledArray
from ..numpy_typing import OnlyArray


class LineOfSightTransform(CoordinateTransform):
    """Coordinate system for data collected along a number of lines-of-sight.

    The first coordinate x1 is the channel number, the second x2 is the
    position along the line-of-sight and ranges from [0, 1].

    Parameters
    ----------
    origin_x
        x positions for the LOS origin in (m).
    origin_y
        y positions for the LOS origin in (m).
    origin_z
        z positions for the LOS origin in (m).
    direction_x
        delta x for the LOS direction in (m).
    direction_y
        delta y for the LOS direction in (m).
    direction_z
        delta z for the LOS direction in (m).
    name
        A string identifier (e.g. the name of the instrument).
    machine_dimensions
        The boundaries of the Tokamak in R-z space in (m):
        ``((Rmin, Rmax), (zmin, zmax)``.
    dl
        The spatial precision along the LOS in (m).
    passes
        Number of passes across the plasma (e.g. interferometer
        with corner cube has passes=2)

    """

    def __init__(
        self,
        origin_x: OnlyArray,
        origin_y: OnlyArray,
        origin_z: OnlyArray,
        direction_x: OnlyArray,
        direction_y: OnlyArray,
        direction_z: OnlyArray,
        name: str = "",
        machine_dimensions: Tuple[Tuple[float, float], Tuple[float, float]] = (
            (1.83, 3.9),
            (-1.75, 2.0),
        ),
        dl: float = 0.01,
        passes: int = 1,
        beamlets: int = 1,
        spot_width: float = 0.0,
        spot_shape: str = "round",
        div_width: float = 0.0,
        # div_h: float = 0.0,
        # spot_height: float = 0.0,
        **kwargs: Any,
    ):

        self.instrument_name: str = name
        self.name = f"{self.instrument_name}_line_of_sight_transform"
        self.x1_name = "channel"
        self.x2_name = "los_position"
        self._machine_dims = machine_dimensions
        self.passes = passes

        self.origin_x = origin_x
        self.origin_y = origin_y
        self.origin_z = origin_z
        self.direction_x = direction_x
        self.direction_y = direction_y
        self.direction_z = direction_z

        # Spot info
        self.spot_width = spot_width
        self.spot_height = spot_width
        self.spot_shape = spot_shape
        self.beamlets = beamlets
        self.div_width = div_width
        self.distribute_beamlets()

        # self.div_height = div_h
        # self.spot_height = spot_height

        # Channel number
        self.x1: list = list(np.arange(0, len(origin_x)))

        # Calculate LOS coordinates
        self.set_dl(dl)

    @property
    def origin(self):
        self._origin = np.array(
            [self.origin_x, self.origin_y, self.origin_z]
        ).transpose()
        return self._origin

    @property
    def direction(self):
        self._direction = np.array(
            [self.direction_x, self.direction_y, self.direction_z]
        ).transpose()
        return self._direction

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, self.__class__):
            return False
        result = self._abstract_equals(other)
        result = cast(bool, result and np.all(self.x_start == other.x_start))
        result = cast(bool, result and np.all(self.z_start == other.z_start))
        result = cast(bool, result and np.all(self.y_start == other.y_start))
        result = cast(bool, result and np.all(self.x_end == other.x_end))
        result = cast(bool, result and np.all(self.z_end == other.z_end))
        result = cast(bool, result and np.all(self.y_end == other.y_end))
        result = cast(bool, result and np.all(self.dl == other.dl))
        result = cast(bool, result and np.all(self.x2 == other.x2))
        result = cast(bool, result and np.all(self.R == other.R))
        result = cast(bool, result and np.all(self.phi == other.phi))
        result = result and self._machine_dims == other._machine_dims
        return result

    def convert_to_xy(
        self, x1: LabeledArray, x2: LabeledArray, t: LabeledArray
    ) -> Coordinates:
        c = np.ceil(x1).astype(int)
        f = np.floor(x1).astype(int)
        x_s = (self.x_start[c] - self.x_start[f]) * (x1 - f) + self.x_start[f]
        x_e = (self.x_end[c] - self.x_end[f]) * (x1 - f) + self.x_end[f]
        y_s = (self.y_start[c] - self.y_start[f]) * (x1 - f) + self.y_start[f]
        y_e = (self.y_end[c] - self.y_end[f]) * (x1 - f) + self.y_end[f]
        x = x_s + (x_e - x_s) * x2
        y = y_s + (y_e - y_s) * x2

        return x, y

    def convert_to_Rz(
        self, x1: LabeledArray, x2: LabeledArray, t: LabeledArray
    ) -> Coordinates:
        """
        Convert to (R,z) the position along the LOS x2 for channel(s) x1.
        """
        c = np.ceil(x1).astype(int)
        f = np.floor(x1).astype(int)
        x_s = (self.x_start[c] - self.x_start[f]) * (x1 - f) + self.x_start[f]
        x_e = (self.x_end[c] - self.x_end[f]) * (x1 - f) + self.x_end[f]
        y_s = (self.y_start[c] - self.y_start[f]) * (x1 - f) + self.y_start[f]
        y_e = (self.y_end[c] - self.y_end[f]) * (x1 - f) + self.y_end[f]
        z_s = (self.z_start[c] - self.z_start[f]) * (x1 - f) + self.z_start[f]
        z_e = (self.z_end[c] - self.z_end[f]) * (x1 - f) + self.z_end[f]
        x = x_s + (x_e - x_s) * x2
        y = y_s + (y_e - y_s) * x2
        z = z_s + (z_e - z_s) * x2

        return np.sqrt(x**2 + y**2), z

    def convert_from_Rz(
        self, R: LabeledArray, z: LabeledArray, t: LabeledArray
    ) -> Coordinates:

        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement a 'convert_from_Rz' "
            "method."
        )

    def distance(
        self,
        direction: str,
        x1: LabeledArray,
        x2: LabeledArray,
        t: LabeledArray,
    ) -> np.ndarray:
        """
        Calculate the physical distances between points along x2.
        """
        x = self.x_start[x1] + (self.x_end[x1] - self.x_start[x1]) * x2
        y = self.y_start[x1] + (self.y_end[x1] - self.y_start[x1]) * x2
        z = self.z_start[x1] + (self.z_end[x1] - self.z_start[x1]) * x2
        spacings = np.sqrt(
            x.diff(direction) ** 2 + z.diff(direction) ** 2 + y.diff(direction) ** 2
        )
        result = zeros_like(x)
        result[{direction: slice(1, None)}] = spacings.cumsum(direction)
        return result

    def distribute_beamlets(self, debug=False):
        """
        Distribute beamlets using information on spot size and divergence.
        TODO: currently working only with spot_width and for circular/round spots
        TODO: expand to include spot_height and other shapes, e.g. rectangular
        """
        # Grid
        n_w = int(np.sqrt(self.beamlets))
        n_v = int(np.sqrt(self.beamlets))
        grid_w = np.linspace(
            -self.spot_width / 2, self.spot_width / 2, n_w * 2 + 1, dtype=float
        )
        grid_w = grid_w[1::2]
        grid_v = np.linspace(
            -self.spot_height / 2, self.spot_height / 2, n_v * 2 + 1, dtype=float
        )
        grid_v = grid_v[1::2]
        W, V = np.meshgrid(grid_w, grid_v)
        self.weightings = np.ones_like(W)

        # Draw spot
        if self.spot_shape == "round":

            th = np.linspace(0.0, 2 * np.pi, 1000)
            spot_w = 0.5 * self.spot_width * np.cos(th)
            spot_y = 0.5 * self.spot_width * np.sin(th)

            if debug:
                plt.figure()
                plt.plot(spot_w, spot_y, "k")
                plt.scatter(W.flatten(), V.flatten(), c="r")
                plt.show()

        else:
            raise ValueError("Spot shape not available")

        # Find distance to virtual focus position
        if self.div_width > 0.0:
            distance = self.spot_width / 2 / np.tan(self.div_width)
        else:
            distance = np.nan

        # Build beamlets
        beamlet_origin_x = np.zeros((len(self.direction_x), self.beamlets))
        beamlet_origin_y = np.zeros((len(self.direction_x), self.beamlets))
        beamlet_origin_z = np.zeros((len(self.direction_x), self.beamlets))
        beamlet_direction_x = np.zeros((len(self.direction_x), self.beamlets))
        beamlet_direction_y = np.zeros((len(self.direction_x), self.beamlets))
        beamlet_direction_z = np.zeros((len(self.direction_x), self.beamlets))
        for i_los in range(len(self.direction_x)):

            dir_x = self.direction_x[i_los]
            dir_y = self.direction_y[i_los]
            dir_z = self.direction_z[i_los]

            orig_x = self.origin_x[i_los]
            orig_y = self.origin_y[i_los]
            orig_z = self.origin_z[i_los]

            normal = np.array([-dir_y, dir_x])
            ang_norm = np.arctan2(normal[1], normal[0])
            # ang_xy = np.arctan2(dir_y, dir_x)

            # Calculate virtual system reference coordinate
            system_x = orig_x - dir_x * distance
            system_y = orig_y - dir_y * distance
            system_z = orig_z - dir_z * distance

            count = 0
            for i_w in range(n_w):
                for i_v in range(n_v):
                    # Move origin along plane normal to the direction
                    beamlet_origin_x[i_los, count] = orig_x + grid_w[i_w] * np.cos(
                        ang_norm
                    )
                    beamlet_origin_y[i_los, count] = orig_y + grid_w[i_w] * np.sin(
                        ang_norm
                    )
                    beamlet_origin_z[i_los, count] = orig_z + grid_v[i_v]

                    # Direction
                    if self.div_width > 0.0:
                        dir_vec = np.array(
                            [
                                beamlet_origin_x[i_los, count] - system_x,
                                beamlet_origin_y[i_los, count] - system_y,
                                beamlet_origin_z[i_los, count] - system_z,
                            ]
                        )
                        dir_vec_norm = dir_vec / np.linalg.norm(dir_vec)
                        beamlet_direction_x[i_los, count] = dir_vec_norm[0]
                        beamlet_direction_y[i_los, count] = dir_vec_norm[1]
                        beamlet_direction_z[i_los, count] = dir_vec_norm[2]
                    else:
                        beamlet_direction_x[i_los, count] = dir_x
                        beamlet_direction_y[i_los, count] = dir_y
                        beamlet_direction_z[i_los, count] = dir_z

                    # TODO: implement divergence, in vertical and horizontal dimensions
                    # Rotate direction
                    # d_direction_x[i_los, count] = (
                    #     grid_w[i_w] * np.sin(self.div_width) / (0.5 * self.spot_width)
                    # )
                    # d_direction_x[i_los, count] = (
                    #     grid_w[i_w] * np.cos(self.div_width) / (0.5 * self.spot_width)
                    # )
                    # ang_xy_new = ang_xy + self.div_width * grid_w[i_w] / (
                    #     0.5 * self.spot_width
                    # )
                    # ang_Rz_new = self.div_width * grid_v[i_v] / \
                    #              (0.5 * self.spot_width)
                    # dir_x_new = np.cos(ang_xy_new)
                    # dir_y_new = np.sin(ang_xy_new)
                    # d_direction_x[i_los, count] = dir_x_new - dir_x
                    # d_direction_y[i_los, count] = dir_y_new - dir_y

                    count += 1

            if debug:
                plt.figure()
                plt.plot(orig_x, orig_y, "kx")
                for i_beamlet in range(self.beamlets):
                    x_ = np.array(
                        [
                            beamlet_origin_x[i_los, i_beamlet],
                            beamlet_origin_x[i_los, i_beamlet]
                            + 1.0 * beamlet_direction_x[i_los, i_beamlet],
                        ]
                    )
                    y_ = np.array(
                        [
                            beamlet_origin_y[i_los, i_beamlet],
                            beamlet_origin_y[i_los, i_beamlet]
                            + 1.0 * beamlet_direction_y[i_los, i_beamlet],
                        ]
                    )
                    plt.plot(x_, y_)
                plt.show()

        self.beamlet_origin_x = beamlet_origin_x
        self.beamlet_origin_y = beamlet_origin_y
        self.beamlet_origin_z = beamlet_origin_z
        self.beamlet_direction_x = beamlet_direction_x
        self.beamlet_direction_y = beamlet_direction_y
        self.beamlet_direction_z = beamlet_direction_z

    # def set_weightings(
    #     self, weightings: LabeledArray,
    # ):
    #     """
    #
    #     Parameters
    #     ----------
    #     weightings
    #
    #     Returns
    #     -------
    #
    #     """
    #
    #     self.weightings = weightings

    def set_dl(
        self,
        dl: float,
    ):
        """
        Set spatial resolutions of the lines of sight, and calculate spatial
        coordinates along the LOS

        Parameters
        ----------
        dl
            Spatial resolution (m)
        """

        if hasattr(self, "rhop"):
            delattr(self, "rhop")

        # Calculate start and end coordinates, R, z and phi for all LOS
        x_start: list = []
        y_start: list = []
        z_start: list = []
        x_end: list = []
        y_end: list = []
        z_end: list = []
        for channel in self.x1:
            x_start.append([])
            y_start.append([])
            z_start.append([])
            x_end.append([])
            y_end.append([])
            z_end.append([])
            for beamlet in range(self.beamlets):
                # origin = (
                #     self.origin_x[channel] + self.delta_origin_x[channel, beamlet],
                #     self.origin_y[channel] + self.delta_origin_y[channel, beamlet],
                #     self.origin_z[channel] + self.delta_origin_z[channel, beamlet],
                # )
                # direction = (
                #     self.direction_x[channel]
                #     + self.delta_direction_x[channel, beamlet],
                #     self.direction_y[channel]
                #     + self.delta_direction_y[channel, beamlet],
                #     self.direction_z[channel]
                #     + self.delta_direction_z[channel, beamlet],
                # )
                origin = (
                    self.beamlet_origin_x[channel, beamlet],
                    self.beamlet_origin_y[channel, beamlet],
                    self.beamlet_origin_z[channel, beamlet],
                )
                direction = (
                    self.beamlet_direction_x[channel, beamlet],
                    self.beamlet_direction_y[channel, beamlet],
                    self.beamlet_direction_z[channel, beamlet],
                )
                _start, _end = find_wall_intersections(
                    origin, direction, machine_dimensions=self._machine_dims
                )
                x_start[channel].append(_start[0])
                y_start[channel].append(_start[1])
                z_start[channel].append(_start[2])
                x_end[channel].append(_end[0])
                y_end[channel].append(_end[1])
                z_end[channel].append(_end[2])

        self.x_start = DataArray(
            x_start,
            coords=[(self.x1_name, self.x1), ("beamlet", np.arange(0, self.beamlets))],
        )
        self.y_start = DataArray(
            y_start,
            coords=[(self.x1_name, self.x1), ("beamlet", np.arange(0, self.beamlets))],
        )
        self.z_start = DataArray(
            z_start,
            coords=[(self.x1_name, self.x1), ("beamlet", np.arange(0, self.beamlets))],
        )

        x_end = DataArray(
            x_end,
            coords=[(self.x1_name, self.x1), ("beamlet", np.arange(0, self.beamlets))],
        )
        y_end = DataArray(
            y_end,
            coords=[(self.x1_name, self.x1), ("beamlet", np.arange(0, self.beamlets))],
        )
        z_end = DataArray(
            z_end,
            coords=[(self.x1_name, self.x1), ("beamlet", np.arange(0, self.beamlets))],
        )

        # Fix identical length of all lines of sight
        los_lengths = np.sqrt(
            (x_end - self.x_start) ** 2
            + (y_end - self.y_start) ** 2
            + (z_end - self.z_start) ** 2
        )
        length = np.max(los_lengths)
        npts = int(np.ceil(length / dl))
        length = float(npts * dl)
        factor = length / los_lengths
        self.x_end = self.x_start + factor * (x_end - self.x_start)
        self.z_end = self.z_start + factor * (z_end - self.z_start)
        self.y_end = self.y_start + factor * (y_end - self.y_start)

        # Calculate coordinates, set to Nan values beyond nominal length
        x: list = []
        y: list = []
        z: list = []
        R: list = []
        phi: list = []
        _x2 = np.linspace(0, 1, npts, dtype=float)
        x2 = DataArray(_x2, coords=[(self.x2_name, _x2)])
        for x1 in self.x1:

            _x, _y = self.convert_to_xy(x1, x2, 0)
            _R, _z = self.convert_to_Rz(x1, x2, 0)
            dist = self.distance(self.x2_name, x1, x2, 0)

            x.append(_x)
            y.append(_y)
            z.append(_z)
            R.append(_R)
            _phi = np.arctan2(_y, _x)
            phi.append(_phi)

        # TODO: Loop over DataArray to NaN where los length > LOS distance
        for x1 in self.x1:

            for beamlet in range(self.beamlets):
                dist = self.distance(self.x2_name, x1, x2, 0)
                _x = x[x1].sel(beamlet=beamlet)
                _y = y[x1].sel(beamlet=beamlet)
                _z = z[x1].sel(beamlet=beamlet)
                _R = R[x1].sel(beamlet=beamlet)
                _phi = phi[x1].sel(beamlet=beamlet)
                x[x1][beamlet] = xr.where(
                    dist[beamlet, :] <= los_lengths[x1, beamlet].values, _x, np.nan
                )
                y[x1][beamlet] = xr.where(
                    dist[beamlet, :] <= los_lengths[x1, beamlet].values, _y, np.nan
                )
                z[x1][beamlet] = xr.where(
                    dist[beamlet, :] <= los_lengths[x1, beamlet].values, _z, np.nan
                )
                R[x1][beamlet] = xr.where(
                    dist[beamlet, :] <= los_lengths[x1, beamlet].values, _R, np.nan
                )
                phi[x1][beamlet] = xr.where(
                    dist[beamlet, :] <= los_lengths[x1, beamlet].values, _phi, np.nan
                )

        # Reset end coordinates to values intersecting the machine walls
        self.x_end = x_end
        self.y_end = y_end
        self.z_end = z_end

        self.x2 = x2
        self.dl = float(dist[0, 1] - dist[0, 0])
        self.x = xr.concat(x, "channel")
        self.y = xr.concat(y, "channel")
        self.z = xr.concat(z, "channel")
        self.phi = xr.concat(phi, "channel")
        self.R = np.sqrt(self.x**2 + self.y**2)
        self.impact_parameter = self.calc_impact_parameter()

    def check_rho_and_profile(
        self, profile_to_map: DataArray, t: LabeledArray = None, calc_rho: bool = False
    ) -> DataArray:
        """
        Check requested time
        """

        time = np.array(t)
        if time.size == 1:
            time = float(time)

        equil_t = self.equilibrium.rhop.t
        equil_ok = (np.min(time) >= np.min(equil_t)) * (np.max(time) <= np.max(equil_t))
        if not equil_ok:
            print(f"Available equilibrium time {np.array(equil_t)}")
            raise ValueError(
                f"Inserted time {time} is not available in Equilibrium object"
            )

        # Make sure rhop.t == requested time
        if not hasattr(self, "rhop") or calc_rho:
            self.convert_to_rho_theta(t=time)
        else:
            if not np.array_equal(self.rhop.t, time):
                self.convert_to_rho_theta(t=time)

        # Check profile
        if not hasattr(profile_to_map, "t"):
            profile = profile_to_map.expand_dims({"t": time})  # type: ignore
        else:
            profile = profile_to_map

        if np.size(time) == 1:
            if np.isclose(profile.t, time, rtol=1.0e-4):
                if "t" in profile_to_map.dims:
                    profile = profile.sel(t=time, method="nearest")
            else:
                raise ValueError("Profile does not include requested time")
        else:
            prof_t = profile.t
            range_ok = (np.min(time) >= np.min(prof_t)) * (
                np.max(time) <= np.max(prof_t)
            )
            if range_ok:
                profile = profile.interp(t=time)
            else:
                raise ValueError("Profile does not include requested time")

        return profile

    def map_profile_to_los(
        self,
        profile_to_map: DataArray,
        t: LabeledArray = None,
        limit_to_sep: bool = True,
        calc_rho: bool = False,
    ) -> DataArray:
        """
        Map profile to lines-of-sight

        Parameters
        ----------
        profile_to_map
            DataArray of the profile to integrate
        t
            Time for interpolation
        limit_to_sep
            Set to True if values outside of separatrix are to be set to 0
        calc_rho
            Calculate rho for specified time-points

        Returns
        -------
            Interpolation of the input profile along the LOS
        """
        self.check_equilibrium()
        profile = self.check_rho_and_profile(profile_to_map, t, calc_rho)

        dims = profile_to_map.dims
        along_los: DataArray
        if "R" in dims and "z" in dims:
            R_ = self.R
            z_ = self.z

            along_los = profile_to_map.interp(R=R_, z=z_).T
        elif "rhop" in dims:
            _rhop = self.rhop
            if "theta" in dims:
                theta_ = self.theta
                along_los = profile.interp(rhop=_rhop, theta=theta_)
            else:
                along_los = profile.interp(rhop=_rhop)

            if limit_to_sep:
                along_los = xr.where(
                    _rhop <= 1,
                    along_los,
                    np.nan,
                )
        else:
            raise NotImplementedError("Coordinates not recognized...")

        drop_dims = [dim for dim in dims if dim != "t"]
        along_los = along_los.drop_vars(drop_dims)
        self.along_los = along_los
        self.profile_to_map = profile_to_map

        return along_los

    def integrate_on_los(
        self,
        profile_to_map: DataArray,
        t: LabeledArray = None,
        limit_to_sep=True,
        calc_rho=False,
        sum_beamlets=True,
    ) -> DataArray:
        """
        Integrate 1D profile along LOS
        Parameters
        ----------
        profile_1d
            DataArray of the 1D profile to integrate
        t
            Time for interpolation
        limit_to_sep
            Set to True if values outside of separatrix are to be set to 0

        Returns
        -------
        Line of sight integral along the LOS
        """
        along_los = self.map_profile_to_los(
            profile_to_map,
            t=t,
            limit_to_sep=limit_to_sep,
            calc_rho=calc_rho,
        )

        if sum_beamlets:
            los_integral = (
                self.passes
                * along_los.sum(["los_position", "beamlet"], skipna=True)
                * self.dl
                / float(self.beamlets)
            )
        else:
            los_integral = (
                self.passes * along_los.sum(["los_position"], skipna=True) * self.dl
            )

        if len(los_integral.channel) == 1:
            los_integral = los_integral.sel(channel=0)

        self.los_integral = los_integral

        return los_integral

    def calc_impact_parameter(self):
        """Calculate the impact parameter in Cartesian space"""
        impact = []
        index = []
        x = []
        y = []
        z = []
        R = []
        for ch in self.x1:
            x_mean = self.x.sel(channel=ch).mean("beamlet")
            y_mean = self.y.sel(channel=ch).mean("beamlet")
            z_mean = self.z.sel(channel=ch).mean("beamlet")
            distance = np.sqrt(x_mean**2 + y_mean**2 + z_mean**2)
            _index = np.unravel_index(distance.argmin(), distance.shape)
            _index_temp = distance.argmin()
            index.append(_index_temp)
            impact.append(distance[_index])
            x.append(x_mean[_index])
            y.append(y_mean[_index])
            z.append(z_mean[_index])
            R.append(np.sqrt(x_mean[_index] ** 2 + y_mean[_index] ** 2))

        impact = Dataset(
            {
                "index": xr.concat(index, "channel"),
                "value": xr.concat(impact, "channel"),
                "x": xr.concat(x, "channel"),
                "y": xr.concat(y, "channel"),
                "z": xr.concat(z, "channel"),
                "R": xr.concat(R, "channel"),
            }
        )

        return impact



"""
Set of dictionaries providing standard names and units for plasma physics quantities
"""

from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

GeneralDataType = str
SpecificDataType = str

#: Structure for type information for :py:class:`xarray.DataArray` objects.
ArrayType = Tuple[Optional[GeneralDataType], Optional[SpecificDataType]]

#: Structure for type information for :py:class:`xarray.Dataset` objects.
DatasetType = Tuple[Optional[SpecificDataType], Dict[str, GeneralDataType]]

DataType = Union[ArrayType, DatasetType]

#: Units expected by each quantity in Indica (key = string identifier)
UNITS: Dict[str, str] = {
    "none": "",
    "number": "#",
    "length": "m",
    "counts": "Counts",
    "arbitrary": "[a.u]",
    "temperature": "eV",
    "density": "$1/m^3$",
    "particle_source": "$1/m^3 s$",
    "density_integrated": "1/$m^2$",
    "brightness": "$W/m^2$",
    "emissivity": "$W/m^3$",
    "intensity": "$count/s$",
    "radiance": "W/m^2/steradian$",
    "spectral_radiance": "W/m^2/steradian/nm$",
    "velocity": "m/s",
    "magnetic_flux": r"$Wb/2\pi$",
    "f": "Wb m",
    "current": "A",
    "current_density": "$A/m^2$",
    "magnetic_field": "T",
    "energy": "J",
    "power": "W",
    "percent": "%",
    "angular_frequency": "rad/s",
    "frequency": "Hz",
    "time": "s",
    "ionisation_rate": "",
    "recombination_rate": "",
    "emission_rate": "W $m^3$",
    "volume_jacobian": "$m^3$",
    "area_jacobian": "$m^2$",
    "volume": "1/$m^3$",
    "area": "$1/m^2$",
    "voltage": "V",
    "pressure_density": "$Pa/m^3$",
    "pressure": "$Pa$",
    "conductivity": r"1/($\Ohm$ m)",
    "wavelength": "nm",
}

# Dictionary of Indica datatypes
#   {"Variable Name" : ("DataArray LongName", "Unit string identifier")}
DATATYPES: Dict[str, Tuple[str, str]] = {
    "dim_0": ("dim_0", "none"),
    "dim_1": ("dim_1", "none"),
    "dim_2": ("dim_2", "none"),
    "index": ("Index", "none"),
    "label": ("Label", "none"),
    "t": ("Time", "time"),
    "time_to_calculate": ("Time", "time"),
    "channel": ("Channel", "number"),
    "wavelength": ("Wavelength", "wavelength"),
    "x": ("x", "length"),
    "y": ("y", "length"),
    "R": ("R", "length"),
    "z": ("z", "length"),
    "R_midplane": ("R-midplane", "length"),
    "z_midplane": ("z-midplane", "length"),
    "R_geo": ("$R_{geo}$", "length"),
    "z_geo": ("$z_{geo}$", "length"),
    "R_mag": ("$R_{mag}$", "length"),
    "z_mag": ("$z_{mag}$", "length"),
    "R_fit": ("R", "length"),
    "z_fit": ("z", "length"),
    "R_hfs": ("$R_{HFS}$", "length"),
    "R_lfs": ("$R_{LFS}$", "length"),
    "R_boundary": ("$R_{boundary}$", "length"),
    "z_boundary": ("$z_{boundary}$", "length"),
    "R_shift": ("$R_{shift}$", "length"),
    "r_minor": ("$r_{min}$", "length"),
    "r_minor_boundary": ("a", "length"),
    "rho": (r"$\rho$", "none"),
    "rhop": (r"$\rho_{pol}$", "none"),
    "rhot": (r"$\rho_{tor}$", "none"),
    "location": ("Location", "length"),  # this is an array (x, y, z)
    "direction": ("Direction", "length"),  # this is an array (dx, dy, dz)
    "element": ("Element", "none"),
    "impurity": ("Impurity", "none"),
    "beamlet": ("Beamlet", "none"),
    "residence_time": (r"Impurity residence time $\tau$", "time"),
    "electron_density": ("$N_e$", "density"),
    "ion_density": ("$N_i$", "density"),
    "fast_ion_density": ("$N_{fast}$", "density"),
    "impurity_density": ("$N_{imp}$", "density"),
    "thermal_neutral_density": ("$N_{neutrals, th}$", "density"),
    "neutral_density": ("$N_{neutrals}$", "density"),
    "electron_density_integrated": (r"$\int N_e$ dl", "density_integrated"),
    "electron_temperature": ("$T_e$", "temperature"),
    "ion_temperature": ("$T_i$", "temperature"),
    "toroidal_rotation": ("$V_{tor}$", "velocity"),
    "toroidal_angular_frequency": (
        r"$\omega_{tor}$",
        "angular_frequency",
    ),
    "centrifugal_asymmetry": ("Centrifugal asymmetry", "none"),
    "centrifugal_asymmetry_multiplier": (
        "Centrifugal asymmetry multiplier",
        "none",
    ),
    "intensity": ("Intensity", "intensity"),
    "radiance": ("Radiance", "radiance"),
    "spectral_radiance": ("Spectral Radiance", "spectral_radiance"),
    "emissivity": ("Emissivity", "emissivity"),
    "total_radiation": ("$P_{rad, tot}$", "emissivity"),
    "sxr_radiation": ("$P_{rad, sxr}$", "emissivity"),
    "total_radiated_power": ("$P_{rad, tot}$", "power"),
    "sxr_radiated_power": ("$P_{rad, sxr}$", "power"),
    "brightness": ("Brightness", "brightness"),
    "line_intensity": ("Line intensity", "brightness"),
    "spectra_raw": ("Spectra", "counts"),
    "spectra": ("Spectra", "brightness"),
    "spectra_fit": ("Spectra (fit)", "brightness"),
    "chi_squared": (r"$\chi^2$", "none"),
    "effective_charge": ("$Z_{eff}$", "none"),
    "mean_charge": ("$<q>$", "none"),
    "fractional_abundance": ("$f_q$", "none"),
    "ion_charge": ("Charge state", "none"),
    "atomic_weight": ("Atomic weight", "none"),
    "atomic_number": ("Atomic number", "none"),
    "element_name": ("Element", "none"),
    "element_symbol": ("Element", "none"),
    "equilibrium_f": ("f", "f"),
    "poloidal_flux": ("Poloidal flux", "magnetic_flux"),
    "poloidal_flux_axis": ("Poloidal flux (axis)", "magnetic_flux"),
    "poloidal_flux_boundary": ("Poloidal flux (separatrix)", "magnetic_flux"),
    "psin": ("Normalised poloidal flux", "magnetic_flux"),
    "toroidal_flux": ("Toroidal flux", "magnetic_flux"),
    "volume_jacobian": ("$V_{jac}$", "volume_jacobian"),
    "area_jacobian": ("$A_{jac}$", "area_jacobian"),
    "volume": ("Volume", "volume"),
    "area": ("Cross-sectional area", "area"),
    "plasma_current": ("$I_{P}$", "current"),
    "toroidal_magnetic_field": ("$B_{T}$", "magnetic_field"),
    "poloidal_magnetic_field": ("$B_{P}$", "magnetic_field"),
    "radial_magnetic_field": ("$B_{r}$", "magnetic_field"),
    "vertical_magnetic_field": ("$B_{z}$", "magnetic_field"),
    "concentration": ("Concentration", "percent"),
    "scd": ("SCD rate coefficient", "ionisation_rate"),
    "acd": ("ACD rate coefficient", "recombination_rate"),
    "ccd": ("CCD rate coefficient", "recombination_rate"),
    "pec": ("PEC coefficient", "emission_rate"),
    "plt": ("PLT coefficient", "emission_rate"),
    "prb": ("PRB coefficient", "emission_rate"),
    "prc": ("PRC coefficient", "emission_rate"),
    "pls": ("PLS coefficient", "emission_rate"),
    "prs": ("PRS coefficient", "emission_rate"),
    "total_radiation_loss_parameter": ("$L_{tot}$", "emission_rate"),
    "sxr_radiation_loss_parameter": ("$L_{sxr}$", "emission_rate"),
    "loop_voltage": ("$V_{loop}$", "voltage"),
    "equilibrium_stored_energy": ("$W_{eq}$", "energy"),
    "thermal_stored_energy": ("$W_{th}$", "energy"),
    "fast_ion_stored_energy": ("$W_{fast}$", "energy"),
    "stored_energy": ("$W_{tot}$", "energy"),
    "ohmic_current_density": ("$j_{ohm}$", "current_density"),
    "bootstrap_current_density": ("$j_{BS}$", "current_density"),
    "nbi_current_density": ("$j_{NBI}$", "current_density"),
    "current_density": ("$j_{tot}$", "current_density"),
    "ohmic_power_density": ("$P_{ohm}$", "power"),
    "nbi_power_density": ("$P_{nbi}$", "power"),
    "nbi_particle_source": ("$S_{nbi}$", "particle_source"),
    "wall_particle_source": ("$S_{wall}$", "particle_source"),
    "particle_source": ("$S_{tot}$", "particle_source"),
    "thermal_pressure": ("$P_{th}", "pressure_density"),
    "electron_pressure": ("$P_{th, e}", "pressure_density"),
    "ion_pressure": ("$P_{th, i}", "pressure_density"),
    "pressure": ("$P_{tot}", "pressure_density"),
    "fast_ion_pressure": ("$P_{fast}", "pressure_density"),
    "parallel_fast_ion_pressure": ("$P_{fast,//}$", "pressure_density"),
    "perpendicular_fast_ion_pressure": (r"$P_{fast,\perp}$", "pressure_density"),
    "thermal_pressure_integral": ("$P_{th}", "pressure"),
    "pressure_integral": ("$P_{tot}", "pressure"),
    "injected_nbi_power": ("$P_{NBI, inj}$", "power"),
    "absorbed_nbi_power": ("$P_{NBI, abs}$", "power"),
    "absorbed_ohmic_power": ("$P_{ohm, abs}$", "power"),
    "safety_factor": ("q", "none"),
    "parallel_conductivity": (r"$\sigma$", "conductivity"),
}



from MDSplus import Float32
from MDSplus import String
from numpy import NaN

# Module for MDSplus helper functions
# Otto Asunta -- 02/03/2018


def getOrCreateNode(t, nodeName, nodeType, nodeHelp, **kwargs):
    """
    ################################################################
    Function gets the node or creates new one if it does not exist.
    ################################################################"""
    try:
        return t.getNode(nodeName)
    except ValueError:
        print("Node " + nodeName + " is created.")
        return createNode(t, nodeName, nodeType, nodeHelp, **kwargs)


def createNode(t, nodeName, nodeType, nodeHelp, **kwargs):
    """################################################################
    Function createNode is used for creating a node and its
    subnode "HELP".

    Input parameters:
    t        = Tree object
    nodeName = name of the node to be added
    nodeType = type of the node to be added
    nodeHelp = help string to be written to .HELP

    Optional parameters:
    dataIn   = data to be written into the node.
               NOTE: Arrays of data need to be defined as array([]),
               i.e. numpy.ndarray, otherwise putData() will fail.
    units    = units of the data (only works if data is given too)
    ################################################################"""

    # Save the current location in the tree
    defaultNode = t.getDefault()
    #
    if "dataIn" in kwargs:
        dataIn = kwargs["dataIn"]
    else:
        dataIn = None

    if "units" in kwargs:
        units = kwargs["units"]
    else:
        units = None

    # Add and prefill the node
    n = t.addNode(nodeName.upper(), nodeType.upper())
    if nodeType.upper() == "SIGNAL":
        if dataIn is None:
            dataIn = t.tdiCompile("build_signal($ROPRAND,*,$ROPRAND)")
            n.putData(dataIn)
        else:
            if units is None:
                n.putData(dataIn)
            else:
                n.putData(dataIn.setUnits(units))

    elif nodeType.upper() == "TEXT":
        if dataIn is None:
            dataIn = " "
            n.putData(dataIn)
        else:
            if units is None:
                n.putData(dataIn)
            else:
                n.putData(String(dataIn).setUnits(units))
    #        d = String(dataIn);

    elif nodeType.upper() == "NUMERIC":
        if dataIn is None:
            dataIn = NaN
            n.putData(dataIn)
        else:
            if units is None:
                n.putData(dataIn)
            else:
                n.putData(Float32(dataIn).setUnits(units))

    #        d = Float32(dataIn);
    elif nodeType.upper() == "STRUCTURE" or nodeType.upper() == "SUBTREE":
        # No more actions needed
        pass
    else:
        #        n = t.setDefault(t.addNode(nodeName,nodeType));
        print("Unknown node type.")
        print("No data filled to node " + nodeName + ".")

    t.setDefault(n)
    #    if(all(~isnan(dataIn))): # If it is not NaN, try to write the data
    # if 'dataIn' in kwargs:
    #     if 'units' in kwargs:
    #         units = kwargs['units'];
    #     else:
    #         units = '';

    #     d.setUnits(units);
    #     print d.units_of()
    #     try:
    #         print 'dataIn=',d.data(), 'units=',d.units
    #         n.putData(d);
    #         #d.setUnits(units);
    #     except:
    #         print "createNode() Failed to write given data into ",n.getFullPath()
    #         print "Continuing regardless"

    #    print t.getDefault()
    t.addNode("HELP", "TEXT").putData(nodeHelp)
    # Return to the point in the tree from which the function was called.
    t.setDefault(defaultNode)
    return t.getNode(nodeName)



from indica.available_quantities import READER_QUANTITIES
from indica.models.abstract_diagnostic import AbstractDiagnostic
from indica.utilities import build_dataarrays
from indica.utilities import check_time_present


class EquilibriumReconstruction(AbstractDiagnostic):
    """Object representing observations from a magnetic reconstruction"""

    def __init__(
        self,
        name: str,
        instrument_method="get_equilibrium",
    ):
        self.name = name
        self.instrument_method = instrument_method
        self.quantities = READER_QUANTITIES[self.instrument_method]

    def _build_bckc_dictionary(self):
        self.bckc = {}
        bckc = {
            "t": self.t,
            "wp": self.wp,
        }
        self.bckc = build_dataarrays(bckc, self.quantities, transform=self.transform)

    def __call__(
        self,
        t=None,
        **kwargs,
    ):
        """Add docs"""
        if self.plasma is None:
            raise ValueError("plasma object is needed")

        if t is None:
            t = self.plasma.time_to_calculate

        check_time_present(t, self.plasma.wp.t)

        self.t = t
        self.wp = self.plasma.wp.sel(t=t)
        self._build_bckc_dictionary()
        return self.bckc



import numpy as np
from scipy.optimize import least_squares

from indica import Plasma


def match_interferometer_los_int(
    models: dict,
    plasma: Plasma,
    data: dict,
    t: float,
    optimise_for: dict = {"smmh1": ["ne"]},
    ne0: float = 5.0e19,
    bounds: tuple = (1.0e17, 1.0e21),
    bckc: dict = {},
):
    """
    Rescale density profiles to match the interferometer measurements
    """

    def residuals(ne0):
        plasma.Ne_prof.y0 = ne0
        plasma.assign_profiles("electron_density", t=t)

        for instrument in optimise_for.keys():
            if instrument in models.keys():
                bckc[instrument] = models[instrument](t=t)

        resid = []
        for instrument in optimise_for.keys():
            if instrument in models.keys():
                for quantity in optimise_for[instrument]:
                    resid.append(
                        data[instrument][quantity].sel(t=t) - bckc[instrument][quantity]
                    )

        return (np.array(resid) ** 2).sum()

    least_squares(residuals, ne0, bounds=bounds, method="dogbox")

    return plasma, bckc



"""Defines an ArrayLike type, using either the new features available
to NumPy (still bleading edge) or a rough approximation. Also provides
a LabeledArray type, which corresponds to Xarray objects with
labeled dimensions and to scalar.

"""

import numbers
import typing
from typing import Optional
from typing import Tuple

import numpy as np
import xarray

try:
    from numpy.typing import ArrayLike
except ImportError:
    ArrayLike = typing.Union[  # type: ignore
        numbers.Number, typing.Sequence[numbers.Number], np.ndarray
    ]

LabeledArray = typing.Union[
    float, int, np.ndarray, xarray.DataArray, xarray.Dataset, xarray.Variable, list
]

OnlyArray = typing.Union[
    np.ndarray,
    xarray.DataArray,
    xarray.Dataset,
    xarray.Variable,
]

OnlyXarray = typing.Union[xarray.DataArray, xarray.Dataset, xarray.Variable]

# Type for generic functions that e.g. return same type as input
XarrayGeneric = typing.TypeVar("XarrayGeneric", xarray.DataArray, xarray.Dataset)

RevisionLike = typing.Union[str, int]

Coordinates = Tuple[LabeledArray, LabeledArray]

OptionalCoordinates = Tuple[Optional[LabeledArray], Optional[LabeledArray]]

FloatOrDataArray = typing.Union[xarray.DataArray, float]



"""
`"Custom accessors"
<https://xarray.pydata.org/en/stable/internals/extending-xarray.html>`_,
are defined to provide additional functionality to
:py:class:`xarray.DataArray` and :py:class:`xarray.Dataset`
objects. These accessors group methods under the namespace
``indica``. E.g.::

"""

from numbers import Number
from typing import Any
from typing import Callable
from typing import cast
from typing import Dict
from typing import Hashable
from typing import List
from typing import Mapping
from typing import Optional
from typing import Tuple
from typing import Union

import numpy as np
from scipy.interpolate import CloughTocher2DInterpolator
from scipy.interpolate import interp1d
from scipy.interpolate import InterpolatedUnivariateSpline
from scipy.interpolate import RectBivariateSpline
import xarray as xr
from xarray.core.utils import either_dict_or_kwargs

from .converters import CoordinateTransform
from .converters.abstractconverter import Coordinates
from .equilibrium import Equilibrium
from .numpy_typing import ArrayLike
from .numpy_typing import LabeledArray


def _convert_coords(
    array: Union[xr.Dataset, xr.DataArray], transform: CoordinateTransform
) -> Coordinates:
    """Convert this array's coordinates to those of the coordinate system
    ``transform``. The result will be cached for future reuse.

    Parameters
    ----------
    array
        The object for which coordinates will be converted.
    transform
        The transform describing the coordinate system to which the
        coordiantes will be converted.

    Returns
    -------
    x1 : LabeledArray
        The first spatial coordinate in the new system
    x2 : LabeledArray
        The second spatial coordinate in the new system

    """
    if transform.x1_name not in array.coords or transform.x2_name not in array.coords:
        self_trans: CoordinateTransform = array.attrs["transform"]
        converter = self_trans.get_converter(transform)
        if converter:
            x1, x2 = converter(
                array.coords[self_trans.x1_name],
                array.coords[self_trans.x2_name],
                array.coords["t"],
            )
        else:
            if "R" not in array.coords or "z" not in array.coords:
                R, z = self_trans.convert_to_Rz(
                    array.coords[self_trans.x1_name],
                    array.coords[self_trans.x2_name],
                    array.coords["t"],
                )
                if "R" not in array.coords:
                    array.coords["R"] = R
                if "z" not in array.coords:
                    array.coords["z"] = z
            else:
                R = array.coords["R"]
                z = array.coords["z"]
            x1, x2 = transform.convert_from_Rz(R, z, array.coords["t"])
        if transform.x1_name not in array.coords:
            array.coords[transform.x1_name] = x1
        if transform.x2_name not in array.coords:
            array.coords[transform.x2_name] = x2
    return array.coords[transform.x1_name], array.coords[transform.x2_name]


@xr.register_dataarray_accessor("indica")
class InDiCAArrayAccessor:
    """Class providing additional functionality to
    :py:class:`xarray.DataArray` objects which is useful for this software.

    """

    def __init__(self, xarray_obj: xr.DataArray):
        self._obj = xarray_obj

    def invert_interp(
        self,
        values: ArrayLike,
        target: str,
        new_dims: Optional[Union[Tuple[str, ...], str]] = None,
        coords: Optional[Mapping[Hashable, ArrayLike]] = None,
        method: str = "linear",
        assume_sorted: bool = False,
        kwargs: Optional[Mapping[str, Any]] = None,
        **coords_kwargs: ArrayLike,
    ) -> xr.DataArray:
        """Performs an inversion to give coordinates at which the data has the
        specified value. Interpolation will be performed with the
        specified method. This approach is efficient if you need to do
        the inversion on a large number of values, but **will only
        work properly if the data is monotonic on the target
        dimension**.

        Parameters
        ----------
        values
            The values of data at which to invert.
        target
            The name of the dimension on which the result contains the
            coordinates at which ``values`` occur.
        new_dims
            The name(s) of the new dimensions introduced by inverting with
            ``values``. Defaults to the name of the original DataArray. If that
            array does not have a name and a named dimension in ``values`` is
            equal to ``target`` then this argument is required.
        coords
            A mapping from dimension names to new coordinates onto
            which to interpolate data. New coordinate can be a scalar,
            array-like or DataArray. If DataArrays are passed as new
            coordinates, their dimensions are used for the
            broadcasting. Missing values are skipped. The target
            dimension **must not be present.**
        method
            {“linear”, “nearest”} for multidimensional array,
            {“linear”, “nearest”, “zero”, “slinear”, “quadratic”,
            “cubic”} for 1-dimensional array. “linear” is used by
            default.
        assume_sorted
            If False, values of coordinates that are interpolated over
            can be in any order and they are sorted first. If True,
            interpolated coordinates are assumed to be an array of
            monotonically increasing values.
        kwargs
            Additional keyword arguments passed to scipy’s
            interpolator. Valid options and their behavior depend on
            if 1-dimensional or multi-dimensional interpolation is
            used.
        coords_kwargs
            The keyword arguments form of coords.

        Returns
        -------
        :
            A DataArray containing the positions on the ``target`` dimension at
            which the original DataArray has ``values``, interpolated as
            necessary.

        """

        def invert_interp_func(data, target_coords, vals):
            not_nan = np.logical_not(np.isnan(data))
            interp = interp1d(data[not_nan], target_coords[not_nan], method, copy=False)
            return interp(vals)

        if coords or coords_kwargs:
            if (coords and target in coords) or target in coords_kwargs:
                raise ValueError(
                    f"Target dimension ('{target}') must not be present in coords."
                )
            interpolated = self._obj.interp(
                coords, method, assume_sorted, kwargs, **coords_kwargs
            )
        else:
            interpolated = self._obj
        if not isinstance(values, xr.DataArray):
            values = xr.DataArray(values)
        new_dim_names: List[str]
        if new_dims is None:
            if values.ndim == 0:
                new_dim_names = []
                value_dims = new_dim_names
            elif target in values.dims:
                if self._obj.name is None:
                    raise ValueError("Values must have a name if 'new_dims' is None.")
                new_dim_names = [
                    cast(str, self._obj.name),
                ]
                value_dims = [
                    target,
                ]
            else:
                new_dim_names = []
                value_dims = new_dim_names
        elif isinstance(new_dims, str):
            new_dim_names = [
                new_dims,
            ]
            value_dims = cast(List, values.dims)
        else:
            new_dim_names = list(new_dims)
            value_dims = cast(List, values.dims)
        data = xr.apply_ufunc(
            invert_interp_func,
            interpolated,
            interpolated.coords[target],
            values,
            input_core_dims=[[target], [target], value_dims],
            output_core_dims=[new_dim_names],
            exclude_dims=set(value_dims),
            vectorize=True,
        )
        data.name = target
        return data

    def invert_root(
        self,
        values: ArrayLike,
        target: str,
        guess: Optional[ArrayLike] = None,
        new_dims: Optional[Union[Tuple[str, ...], str]] = None,
        coords: Optional[Mapping[Hashable, ArrayLike]] = None,
        method: str = "linear",
        assume_sorted: bool = False,
        kwargs: Optional[Mapping[str, Any]] = None,
        **coords_kwargs: ArrayLike,
    ) -> xr.DataArray:
        """Performs an inversion to give coordinates at which the data has the
        specified value. Cubic spline interpolation will be performed with the
        specified method and then the root(s) of the interpolant will
        be found. If there are multiple roots then a guess must be
        provided to determine which one to use.

        Parameters
        ----------
        values
            The values of data at which to invert.
        target
            The name of the dimension on which the result contains the
            coordinates at which ``values`` occur.
        guess
            An estimate of the values being solved for. If there are multiple
            roots then the one closest to the guess will be used.
        new_dims
            The name(s) of the new dimensions introduced by inverting with
            ``values``. Defaults to the name of the original DataArray. If that
            array does not have a name and a named dimension in ``values`` is
            equal to ``target`` then this argument is required.
        coords
            A mapping from dimension names to new coordinates onto
            which to interpolate data. New coordinate can be a scalar,
            array-like or DataArray. If DataArrays are passed as new
            coordinates, their dimensions are used for the
            broadcasting. Missing values are skipped. The target
            dimension **must not be present.**
        method
            {“linear”, “nearest”} for multidimensional array,
            {“linear”, “nearest”, “zero”, “slinear”, “quadratic”,
            “cubic”} for 1-dimensional array. “linear” is used by
            default. Only applies for interpolation onto ``coords``.
        assume_sorted
            If False, values of coordinates that are interpolated over
            can be in any order and they are sorted first. If True,
            interpolated coordinates are assumed to be an array of
            monotonically increasing values.
        kwargs
            Additional keyword arguments passed to scipy’s
            interpolator. Valid options and their behavior depend on
            if 1-dimensional or multi-dimensional interpolation is
            used.
        coords_kwargs
            The keyword arguments form of coords.

        Returns
        -------
        :
            A DataArray containing the positions on the ``target`` dimension at
            which the original DataArray has ``values``, interpolated as
            necessary.

        """

        def invert_root_func(data, target_coords, value, local_guess):
            not_nan = np.logical_not(np.isnan(data))
            interp = InterpolatedUnivariateSpline(
                target_coords[not_nan], (data - value)[not_nan], ext=2
            )
            roots = interp.roots()
            if len(roots) == 0:
                start = target_coords.argmin()
                end = target_coords.argmax()
                if np.isclose(data[start], value):
                    roots = np.concatenate((roots, [target_coords[start]]))
                if np.isclose(data[end], value):
                    roots = np.concatenate((roots, [target_coords[end]]))
                if len(roots) == 0:
                    raise ValueError(f"Provided data is not available at {value}.")
            elif len(roots) > 1 and guess is None:
                raise ValueError(
                    "A guess must be provided when there is more than one root."
                )
            i = np.abs(roots - local_guess).argmin()
            return roots[i]

        if coords or coords_kwargs:
            if (coords and target in coords) or target in coords_kwargs:
                raise ValueError(
                    f"Target dimension ('{target}') must not be present in coords."
                )
            interpolated = self._obj.interp(
                coords, method, assume_sorted, kwargs, **coords_kwargs
            )
        else:
            interpolated = self._obj
        if not isinstance(values, xr.DataArray):
            values = xr.DataArray(values)
        new_dim_names: List[str]
        value_dims: List[str]
        if new_dims is None:
            if values.ndim == 0:
                new_dim_names = []
                value_dims = new_dim_names
            elif target in values.dims:
                if values.ndim == 1:
                    if self._obj.name is None:
                        raise ValueError(
                            "Values must have a name if 'new_dims' is None."
                        )
                    new_dim_names = [
                        cast(str, self._obj.name),
                    ]
                    value_dims = [
                        target,
                    ]
                else:
                    raise ValueError(
                        f"Target dimension '{target}' can only be used in `values` "
                        "if `values` is a 1-D array."
                    )
            else:
                new_dim_names = []
                value_dims = new_dim_names
        elif isinstance(new_dims, str):
            new_dim_names = [
                new_dims,
            ]
        else:
            new_dim_names = list(new_dims)

        # FIXME: Still need to handling of guess dimensions...
        if isinstance(guess, xr.DataArray) and target in guess.dims:
            guess_core_dims = [target]
        else:
            guess_core_dims = []
        if isinstance(values, xr.DataArray):
            value_dims = [cast(str, dim) for dim in values.dims if dim in new_dim_names]
        else:
            value_dims = []
        data = xr.apply_ufunc(
            invert_root_func,
            interpolated,
            interpolated.coords[target],
            values,
            0.0 if guess is None else guess,
            input_core_dims=[[target], [target], value_dims, guess_core_dims],
            output_core_dims=[new_dim_names],
            exclude_dims=set((target,)),
            vectorize=True,
        )
        data.name = target
        return data

    @staticmethod
    def _get_unlabeled_interpolation_1d(
        assume_sorted: bool, degree: int
    ) -> Callable[[ArrayLike, ArrayLike, ArrayLike, Optional[ArrayLike]], ArrayLike]:
        def interpolate_1d(x, z, x_interp, x_zero):
            if assume_sorted and x_zero is None:
                interpolant = InterpolatedUnivariateSpline(x, z, k=degree)
                domain = (x[0], x[-1])
            else:
                if x_zero is not None and x_zero not in x:
                    x = np.append(x, x_zero)
                    z = np.append(z, 0.0)
                order = np.argsort(x)
                interpolant = InterpolatedUnivariateSpline(x[order], z[order], k=degree)
                domain = (x[order[0]], x[order[-1]])
            result = interpolant(x_interp)
            result[np.logical_or(x_interp < domain[0], x_interp > domain[1])] = float(
                "nan"
            )
            return result

        return interpolate_1d

    @staticmethod
    def _get_unlabeled_interpolation_2d(
        assume_sorted: bool, degree: int
    ) -> Callable[
        [
            ArrayLike,
            ArrayLike,
            ArrayLike,
            ArrayLike,
            ArrayLike,
            Optional[ArrayLike],
            Optional[ArrayLike],
        ],
        ArrayLike,
    ]:
        def interpolate_2d(x, y, z, x_interp, y_interp, x_zero, y_zero):
            if assume_sorted:
                interpolant = RectBivariateSpline(x, y, z, kx=degree, ky=degree)
                result = interpolant(x_interp, y_interp)
                domain = ((x[0], x[-1]), (y[0], y[-1]))
            else:
                xorder = np.argsort(x)
                yorder = np.argsort(y)
                z_ordered = np.take(np.take(z, xorder, 0), yorder, 1)
                interpolant = RectBivariateSpline(
                    np.take(x, xorder),
                    np.take(y, yorder),
                    z_ordered,
                    kx=degree,
                    ky=degree,
                )
                new_xorder = np.argsort(x_interp)
                new_yorder = np.argsort(y_interp)
                vals = interpolant(
                    np.take(x_interp, new_xorder),
                    np.take(y_interp, new_yorder),
                ).T
                vals2 = np.empty_like(vals)
                result = np.empty_like(vals)
                np.put(vals2, new_xorder, vals, 1)
                np.put(result, new_yorder, vals2, 0)
                domain = ((x[xorder[0]], x[xorder[-1]]), (y[yorder[0]], y[yorder[-1]]))
            if (
                x_zero is not None
                and y_zero is not None
                and (x_zero not in x or y_zero not in y)
            ):
                x_offsets = x - x_zero
                y_offsets = y - y_zero
                x_below_zero_i = np.argmax(
                    np.where(x_offsets < 0, x_offsets, float("-inf"))
                )
                x_above_zero_i = np.argmin(
                    np.where(x_offsets > 0, x_offsets, float("inf"))
                )
                y_below_zero_i = np.argmax(
                    np.where(y_offsets < 0, y_offsets, float("-inf"))
                )
                y_above_zero_i = np.argmin(
                    np.where(y_offsets > 0, y_offsets, float("inf"))
                )
                x_below_zero = x[x_below_zero_i]
                x_above_zero = x[x_above_zero_i]
                y_below_zero = y[y_below_zero_i]
                y_above_zero = y[y_above_zero_i]
                xmask = np.logical_and(
                    x_interp >= x_below_zero, x_interp <= x_above_zero
                )
                ymask = np.logical_and(
                    y_interp >= y_below_zero, y_interp <= y_above_zero
                )
                if isinstance(xmask, np.ndarray) and isinstance(ymask, np.ndarray):
                    xmask = np.expand_dims(xmask, 1)
                    x_interp = np.expand_dims(x_interp, 1)
                mask = np.logical_and(xmask, ymask)
                if np.any(mask):
                    interpolant = CloughTocher2DInterpolator(
                        np.array(
                            [
                                [x_below_zero, y_below_zero],
                                [x_below_zero, y_above_zero],
                                [x_above_zero, y_below_zero],
                                [x_above_zero, y_above_zero],
                                [x_zero, y_zero],
                            ]
                        ),
                        np.array(
                            [
                                z[x_below_zero_i, y_below_zero_i],
                                z[x_below_zero_i, y_above_zero_i],
                                z[x_above_zero_i, y_below_zero_i],
                                z[x_above_zero_i, y_above_zero_i],
                                0.0,
                            ]
                        ),
                    )
                    xx_interp, yy_interp = np.broadcast_arrays(x_interp, y_interp)
                    result[mask] = interpolant(
                        np.stack([xx_interp[mask], yy_interp[mask]], axis=-1)
                    )
            out_of_domain = np.logical_or(
                np.logical_or(x_interp < domain[0][0], x_interp > domain[0][1]),
                np.logical_or(y_interp < domain[1][0], y_interp > domain[1][1]),
            )
            result[out_of_domain] = float("nan")
            return result

        return interpolate_2d

    def interp2d(
        self,
        coords: Optional[Mapping[Hashable, ArrayLike]] = None,
        zero_coords: Optional[Mapping[Hashable, ArrayLike]] = None,
        method: Union[str, int] = "linear",
        assume_sorted: bool = False,
        **coords_kwargs: ArrayLike,
    ) -> xr.DataArray:
        """Performs interpolation along two dimensions simultatiously. Unlike
        :py:meth:`xarray.DataArray.interp`, this routine supports
        higher-order interpolation than linear by usinmg
        :py:func:`scipy.interpolate.RectBivariateSpline` (although
        with potentially-inefficient vectorisation). All options are the same
        as in the xarray method. However, interpolation will not be performed
        on any of the non-dimensional coordinates, unlike in the xarray method.

        Parameters
        ----------
        coords
            A mapping from dimension names to new coordinates onto
            which to interpolate data. New coordinate can be a scalar,
            array-like or DataArray. If DataArrays are passed as new
            coordinates, their dimensions are used for the
            broadcasting. Missing values are skipped.
        zero_coords
            Coordinates for any points where the data is known to be equal to
            zero but which are not captured by the grid.
        method
            A string {“linear”, “nearest”, “quadratic”, “cubic”} or an integer
            indicating the degree of the splines.
        assume_sorted
            If False, values of coordinates that are interpolated over
            can be in any order and they are sorted first. If True,
            interpolated coordinates are assumed to be an array of
            monotonically increasing values.
        coords_kwargs
            The keyword arguments form of coords.

        Returns
        -------
        :
            A DataArray containing the positions on the ``target`` dimension at
            which the original DataArray has ``values``, interpolated as
            necessary.

        """
        METHODS = {"nearest": 0, "linear": 1, "quadratic": 2, "cubic": 3}
        if isinstance(method, str):
            degree = METHODS[method]
        else:
            degree = method

        _coords = either_dict_or_kwargs(coords, coords_kwargs, "indica.interp2d")
        if len(_coords) > 2:
            raise ValueError("Must not pass more than 2 coordinates to interp2d.")
        if zero_coords and len(zero_coords) != len(_coords):
            raise ValueError(
                "Argument zero_coords must contain the same number of entries "
                "as coordinates on which interpolation is being performed."
            )
        ordered_coords: List[Tuple[str, Union[xr.DataArray, Number]]] = []
        output_core: List[str] = []
        interp_core: List[List[str]] = []
        rename_dims: Dict[str, str] = {}
        ordered_zero_coords: List[ArrayLike] = []
        for k, v in _coords.items():
            if isinstance(v, (np.ndarray, list, tuple)):
                new_dim = "__new_" + cast(str, k)
                ordered_coords.append((cast(str, k), xr.DataArray(v, dims=new_dim)))
                output_core.append(new_dim)
                interp_core.append([new_dim])
                rename_dims[new_dim] = cast(str, k)
            elif isinstance(v, xr.DataArray):
                if k in v.dims:
                    new_dim = "__new_" + cast(str, k)
                    interp_core.append([new_dim])
                    rename_dims[new_dim] = cast(str, k)
                    if new_dim not in output_core:
                        output_core.append(new_dim)
                    ordered_coords.append((cast(str, k), v.rename({k: new_dim})))
                else:
                    interp_core.append([])
                    ordered_coords.append((cast(str, k), v))
            else:
                # Not really sure what mypy is complaining about here.
                ordered_coords.append((k, v))  # type: ignore
                # mypy doesn't allow empty append.
                output_core.append([])  # type: ignore
            if zero_coords:
                ordered_zero_coords.append(zero_coords[k])
            else:
                ordered_zero_coords.append(None)  # type: ignore
        if len(_coords) > 1:
            input_core: List[List[str]] = [
                [ordered_coords[0][0]],
                [ordered_coords[1][0]],
                list(cast(Mapping[str, Any], _coords)),
            ]
            result = xr.apply_ufunc(
                self._get_unlabeled_interpolation_2d(assume_sorted, degree),
                self._obj.coords[ordered_coords[0][0]],
                self._obj.coords[ordered_coords[1][0]],
                self._obj,
                ordered_coords[0][1],
                ordered_coords[1][1],
                ordered_zero_coords[0],
                ordered_zero_coords[1],
                input_core_dims=input_core
                + interp_core
                + cast(List[List[str]], [[], []]),
                output_core_dims=[output_core],
                exclude_dims=set(_coords),
                vectorize=True,
            )
        else:
            input_core = [
                [ordered_coords[0][0]],
                list(cast(Mapping[str, Any], _coords)),
            ]
            result = xr.apply_ufunc(
                self._get_unlabeled_interpolation_1d(assume_sorted, degree),
                self._obj.coords[ordered_coords[0][0]],
                self._obj,
                ordered_coords[0][1],
                ordered_zero_coords[0],
                input_core_dims=input_core + interp_core + cast(List[List[str]], [[]]),
                output_core_dims=[output_core],
                exclude_dims=set(_coords),
                vectorize=True,
            )
        if len(rename_dims) > 0:
            result = result.rename(rename_dims)
        for name, coord in ordered_coords:
            if name not in result.coords:
                result.coords[name] = coord
        return result

    def convert_coords(self, transform: CoordinateTransform) -> Coordinates:
        """Convert this array's coordinates to those of the coordinate system
        ``transform``. The result will be cached for future reuse.

        Parameters
        ----------
        transform
            The transform describing the coordinate system to which the
            coordiantes will be converted.

        Returns
        -------
        x1 : LabeledArray
            The first spatial coordinate in the new system
        x2 : LabeledArray
            The second spatial coordinate in the new system

        """
        return _convert_coords(self._obj, transform)

    def get_coords(
        self, transform: Optional[CoordinateTransform] = None
    ) -> Tuple[LabeledArray, LabeledArray, LabeledArray]:
        """Get this array's coordinates, including time, in the coordinate
        system ``transform``. The result will be cached for future
        reuse.

        Parameters
        ----------
        transform
            The transform describing the coordinate system to which the
            coordiantes will be converted.

        Returns
        -------
        x1 : LabeledArray
            The first spatial coordinate in the new system
        x2 : LabeledArray
            The second spatial coordinate in the new system
        t : LabeledArray
            The time coordinate

        """
        if transform is None:
            transform = self._obj.attrs["transform"]
        return self.convert_coords(transform) + (self._obj.coords["t"],)

    @property
    def equilibrium(self) -> Optional[Equilibrium]:
        """The equilibrium object currently used by this DataArray (or, more
        accurately, by its
        :py:class:`~indica.converters.CoordinateTransform`
        object).

        """
        if "transform" in self._obj.attrs:
            return getattr(self._obj.attrs["transform"], "equilibrium", None)
        else:
            return None

    @equilibrium.setter
    def equilibrium(self, value: Equilibrium):
        if value == self.equilibrium:
            return
        self._obj.attrs["transform"].set_equilibrium(value)

    @equilibrium.deleter
    def equilibrium(self):
        if hasattr(self._obj.attrs["transform"], "equilibrium"):
            del self._obj.attrs["transform"].equilibrium


@xr.register_dataset_accessor("indica")
class InDiCADatasetAccessor:
    """Class providing additional functionality to
    :py:class:`xarray.Dataset` objects which is useful for this software.

    """

    def __init__(self, xarray_obj: xr.Dataset):
        self._obj = xarray_obj

    def convert_coords(self, transform: CoordinateTransform) -> Coordinates:
        """Convert this dataset's coordinates to those of the coordinate system
        ``transform``. The result will be cached for future reuse.

        Parameters
        ----------
        transform
            The transform describing the coordinate system to which the
            coordiantes will be converted.

        Returns
        -------
        x1 : LabeledArray
            The first spatial coordinate in the new system
        x2 : LabeledArray
            The second spatial coordinate in the new system

        """
        return _convert_coords(self._obj, transform)

    def get_coords(
        self, transform: Optional[CoordinateTransform] = None
    ) -> Tuple[LabeledArray, LabeledArray, LabeledArray]:
        """Get this dataset's coordinates, including time, in the coordinate
        system ``transform``. The result will be cached for future
        reuse.

        Parameters
        ----------
        transform
            The transform describing the coordinate system to which the
            coordiantes will be converted.

        Returns
        -------
        x1 : LabeledArray
            The first spatial coordinate in the new system
        x2 : LabeledArray
            The second spatial coordinate in the new system
        t : LabeledArray
            The time coordinate

        """
        if transform is None:
            transform = self._obj.attrs["transform"]
        return self.convert_coords(transform) + (self._obj.coords["t"],)



from copy import deepcopy

import numpy as np
from xarray import DataArray

from indica import Equilibrium
from indica.numpy_typing import LabeledArray
import indica.physics as ph
from indica.utilities import get_element_info


def centrifugal_asymmetry_parameter(
    ion_density: DataArray,
    ion_temperature: DataArray,
    electron_temperature: DataArray,
    toroidal_rotation: DataArray,
    meanz: DataArray,
    zeff: DataArray,
    main_ion: str,
):
    """
    Indica-native wrapper to calculate the centrifugal asymmetry parameter a-la Wesson
    """

    elements = ion_density.element
    zeff = zeff.sum("element")

    asymmetry_parameter = deepcopy(ion_density)
    for elem in elements.data:
        _elem = str(elem)
        main_ion_mass = get_element_info(main_ion)[1]
        mass = get_element_info(_elem)[1]
        asymmetry_parameter.loc[dict(element=_elem)] = ph.centrifugal_asymmetry(
            ion_temperature,
            electron_temperature,
            mass,
            meanz.sel(element=_elem).drop_vars("element"),
            zeff,
            main_ion_mass,
            toroidal_rotation=toroidal_rotation,
        )

    return asymmetry_parameter


def centrifugal_asymmetry_2d_map(
    profile_to_map: DataArray,
    asymmetry_parameter: DataArray,
    equilibrium: Equilibrium,
    t: LabeledArray = None,
):
    """Map centrifugal asymmetric profiles to 2D"""

    if t is None:
        t = profile_to_map.t.values

    if "t" in profile_to_map.dims:
        _profile_to_map = profile_to_map.interp(t=t)
    else:
        _profile_to_map = profile_to_map
    if "t" in asymmetry_parameter.dims:
        _asymmetry_parameter = asymmetry_parameter.interp(t=t)
    else:
        _asymmetry_parameter = asymmetry_parameter

    rho_2d = equilibrium.rhop.interp(t=t)
    R_0 = equilibrium.rmjo.interp(t=t).interp(rhop=rho_2d).drop_vars("rhop")

    _profile_2d = _profile_to_map.interp(rhop=rho_2d).drop_vars("rhop")
    profile_2d = _profile_2d * np.exp(
        _asymmetry_parameter.interp(rhop=rho_2d) * (rho_2d.R**2 - R_0**2)
    )

    return profile_2d
